#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from docx import Document
import re
import pandas as pd
from collections import Counter
from trustrag.modules.document import rag_tokenizer
from io import BytesIO


class DocxParser:

    def __extract_table_content(self, tb):
        df = []
        for row in tb.rows:
            df.append([c.text for c in row.cells])
        return self.__compose_table_content(pd.DataFrame(df))

    def __compose_table_content(self, df):

        def blockType(b):
            patt = [
                ("^(20|19)[0-9]{2}[年/-][0-9]{1,2}[月/-][0-9]{1,2}日*$", "Dt"),
                (r"^(20|19)[0-9]{2}年$", "Dt"),
                (r"^(20|19)[0-9]{2}[年/-][0-9]{1,2}月*$", "Dt"),
                ("^[0-9]{1,2}[月/-][0-9]{1,2}日*$", "Dt"),
                (r"^第*[一二三四1-4]季度$", "Dt"),
                (r"^(20|19)[0-9]{2}年*[一二三四1-4]季度$", "Dt"),
                (r"^(20|19)[0-9]{2}[ABCDE]$", "DT"),
                ("^[0-9.,+%/ -]+$", "Nu"),
                (r"^[0-9A-Z/\._~-]+$", "Ca"),
                (r"^[A-Z]*[a-z' -]+$", "En"),
                (r"^[0-9.,+-]+[0-9A-Za-z/$￥%<>（）()' -]+$", "NE"),
                (r"^.{1}$", "Sg")
            ]
            for p, n in patt:
                if re.search(p, b):
                    return n
            tks = [t for t in rag_tokenizer.tokenize(b).split(" ") if len(t) > 1]
            if len(tks) > 3:
                if len(tks) < 12:
                    return "Tx"
                else:
                    return "Lx"

            if len(tks) == 1 and rag_tokenizer.tag(tks[0]) == "nr":
                return "Nr"

            return "Ot"

        if len(df) < 2:
            return []
        max_type = Counter([blockType(str(df.iloc[i, j])) for i in range(
            1, len(df)) for j in range(len(df.iloc[i, :]))])
        max_type = max(max_type.items(), key=lambda x: x[1])[0]

        colnm = len(df.iloc[0, :])
        hdrows = [0]  # header is not nessesarily appear in the first line
        if max_type == "Nu":
            for r in range(1, len(df)):
                tys = Counter([blockType(str(df.iloc[r, j]))
                              for j in range(len(df.iloc[r, :]))])
                tys = max(tys.items(), key=lambda x: x[1])[0]
                if tys != max_type:
                    hdrows.append(r)

        lines = []
        for i in range(1, len(df)):
            if i in hdrows:
                continue
            hr = [r - i for r in hdrows]
            hr = [r for r in hr if r < 0]
            t = len(hr) - 1
            while t > 0:
                if hr[t] - hr[t - 1] > 1:
                    hr = hr[t:]
                    break
                t -= 1
            headers = []
            for j in range(len(df.iloc[i, :])):
                t = []
                for h in hr:
                    x = str(df.iloc[i + h, j]).strip()
                    if x in t:
                        continue
                    t.append(x)
                t = ",".join(t)
                if t:
                    t += ": "
                headers.append(t)
            cells = []
            for j in range(len(df.iloc[i, :])):
                if not str(df.iloc[i, j]):
                    continue
                cells.append(headers[j] + str(df.iloc[i, j]))
            lines.append(";".join(cells))

        if colnm > 3:
            return lines
        return ["\n".join(lines)]

    def parse(self, fnm, from_page=0, to_page=100000):
        self.doc = Document(fnm) if isinstance(
            fnm, str) else Document(BytesIO(fnm))
        pn = 0
        secs = []
        for p in self.doc.paragraphs:
            if pn > to_page:
                break
            if from_page <= pn < to_page and p.text.strip():
                secs.append((p.text, p.style.name))
            for run in p.runs:
                if 'lastRenderedPageBreak' in run._element.xml:
                    pn += 1
                    continue
                if 'w:br' in run._element.xml and 'type="page"' in run._element.xml:
                    pn += 1

        tbls = [self.__extract_table_content(tb) for tb in self.doc.tables]
        secs=[sec[0] for sec in secs]
        return secs+tbls
if __name__ == '__main__':
    # dp=DocxParser()
    # secs, tbls=dp.parse('/data/users/searchgpt/yq/GoMate/data/docs/夏至各地习俗.docx')
    # print(secs)
    # print(tbls)

    dp = DocxParser()
    secs, tbls = dp.parse('/data/users/searchgpt/yq/GoMate_dev/data/docs/计算所现行制度汇编202406/人力资源处/00-计算所工作人员年休假实施办法-发文.doc')
    print(secs)
    print(tbls)



# #  Licensed under the Apache License, Version 2.0 (the "License");
# #  you may not use this file except in compliance with the License.
# #  You may obtain a copy of the License at
# #
# #      http://www.apache.org/licenses/LICENSE-2.0
# #
# #  Unless required by applicable law or agreed to in writing, software
# #  distributed under the License is distributed on an "AS IS" BASIS,
# #  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# #  See the License for the specific language governing permissions and
# #  limitations under the License.
# #
# 
# import os
# import random
# 
# import xgboost as xgb
# from io import BytesIO
# import torch
# import re
# import pdfplumber
# import logging
# from PIL import Image, ImageDraw
# import numpy as np
# from timeit import default_timer as timer
# from PyPDF2 import PdfReader as pdf2_read
# 
# # from trustrag.modules.document.utils import get_project_base_directory
# # from trustrag.modules.document.vision import OCR, Recognizer, LayoutRecognizer, TableStructureRecognizer
# from trustrag.modules.document import rag_tokenizer
# from copy import deepcopy
# # from huggingface_hub import snapshot_download
# 
# logging.getLogger("pdfminer").setLevel(logging.WARNING)
# 
# 
# class PdfParser:
#     def __init__(self):
#         self.ocr = OCR()
#         if hasattr(self, "model_speciess"):
#             self.layouter = LayoutRecognizer("layout." + self.model_speciess)
#         else:
#             self.layouter = LayoutRecognizer("layout")
#         self.tbl_det = TableStructureRecognizer()
# 
#         self.updown_cnt_mdl = xgb.Booster()
#         if torch.cuda.is_available():
#             self.updown_cnt_mdl.set_param({"device": "cuda"})
#         try:
#             model_dir = os.path.join(
#                 get_project_base_directory(),
#                 "rag/res/deepdoc")
#             self.updown_cnt_mdl.load_model(os.path.join(
#                 model_dir, "updown_concat_xgb.model"))
#         except Exception as e:
#             model_dir = snapshot_download(
#                 repo_id="InfiniFlow/text_concat_xgb_v1.0",
#                 local_dir=os.path.join(get_project_base_directory(), "rag/res/deepdoc"),
#                 local_dir_use_symlinks=False)
#             self.updown_cnt_mdl.load_model(os.path.join(
#                 model_dir, "updown_concat_xgb.model"))
# 
#         self.page_from = 0
#         """
#         If you have trouble downloading HuggingFace models, -_^ this might help!!
# 
#         For Linux:
#         export HF_ENDPOINT=https://hf-mirror.com
# 
#         For Windows:
#         Good luck
#         ^_-
# 
#         """
# 
#     def __char_width(self, c):
#         return (c["x1"] - c["x0"]) // max(len(c["text"]), 1)
# 
#     def __height(self, c):
#         return c["bottom"] - c["top"]
# 
#     def _x_dis(self, a, b):
#         return min(abs(a["x1"] - b["x0"]), abs(a["x0"] - b["x1"]),
#                    abs(a["x0"] + a["x1"] - b["x0"] - b["x1"]) / 2)
# 
#     def _y_dis(
#             self, a, b):
#         return (
#                        b["top"] + b["bottom"] - a["top"] - a["bottom"]) / 2
# 
#     def _match_proj(self, b):
#         proj_patt = [
#             r"第[零一二三四五六七八九十百]+章",
#             r"第[零一二三四五六七八九十百]+[条节]",
#             r"[零一二三四五六七八九十百]+[、是 　]",
#             r"[\(（][零一二三四五六七八九十百]+[）\)]",
#             r"[\(（][0-9]+[）\)]",
#             r"[0-9]+(、|\.[　 ]|）|\.[^0-9./a-zA-Z_%><-]{4,})",
#             r"[0-9]+\.[0-9.]+(、|\.[ 　])",
#             r"[⚫•➢①② ]",
#         ]
#         return any([re.match(p, b["text"]) for p in proj_patt])
# 
#     def _updown_concat_features(self, up, down):
#         w = max(self.__char_width(up), self.__char_width(down))
#         h = max(self.__height(up), self.__height(down))
#         y_dis = self._y_dis(up, down)
#         LEN = 6
#         tks_down = rag_tokenizer.tokenize(down["text"][:LEN]).split(" ")
#         tks_up = rag_tokenizer.tokenize(up["text"][-LEN:]).split(" ")
#         tks_all = up["text"][-LEN:].strip() \
#                   + (" " if re.match(r"[a-zA-Z0-9]+",
#                                      up["text"][-1] + down["text"][0]) else "") \
#                   + down["text"][:LEN].strip()
#         tks_all = rag_tokenizer.tokenize(tks_all).split(" ")
#         fea = [
#             up.get("R", -1) == down.get("R", -1),
#             y_dis / h,
#             down["page_number"] - up["page_number"],
#             up["layout_type"] == down["layout_type"],
#             up["layout_type"] == "text",
#             down["layout_type"] == "text",
#             up["layout_type"] == "table",
#             down["layout_type"] == "table",
#             True if re.search(
#                 r"([。？！；!?;+)）]|[a-z]\.)$",
#                 up["text"]) else False,
#             True if re.search(r"[，：‘“、0-9（+-]$", up["text"]) else False,
#             True if re.search(
#                 r"(^.?[/,?;:\]，。；：’”？！》】）-])",
#                 down["text"]) else False,
#             True if re.match(r"[\(（][^\(\)（）]+[）\)]$", up["text"]) else False,
#             True if re.search(r"[，,][^。.]+$", up["text"]) else False,
#             True if re.search(r"[，,][^。.]+$", up["text"]) else False,
#             True if re.search(r"[\(（][^\)）]+$", up["text"])
#                     and re.search(r"[\)）]", down["text"]) else False,
#             self._match_proj(down),
#             True if re.match(r"[A-Z]", down["text"]) else False,
#             True if re.match(r"[A-Z]", up["text"][-1]) else False,
#             True if re.match(r"[a-z0-9]", up["text"][-1]) else False,
#             True if re.match(r"[0-9.%,-]+$", down["text"]) else False,
#             up["text"].strip()[-2:] == down["text"].strip()[-2:] if len(up["text"].strip()
#                                                                         ) > 1 and len(
#                 down["text"].strip()) > 1 else False,
#             up["x0"] > down["x1"],
#             abs(self.__height(up) - self.__height(down)) / min(self.__height(up),
#                                                                self.__height(down)),
#             self._x_dis(up, down) / max(w, 0.000001),
#             (len(up["text"]) - len(down["text"])) /
#             max(len(up["text"]), len(down["text"])),
#             len(tks_all) - len(tks_up) - len(tks_down),
#             len(tks_down) - len(tks_up),
#             tks_down[-1] == tks_up[-1],
#             max(down["in_row"], up["in_row"]),
#             abs(down["in_row"] - up["in_row"]),
#             len(tks_down) == 1 and rag_tokenizer.tag(tks_down[0]).find("n") >= 0,
#             len(tks_up) == 1 and rag_tokenizer.tag(tks_up[0]).find("n") >= 0
#         ]
#         return fea
# 
#     @staticmethod
#     def sort_X_by_page(arr, threashold):
#         # sort using y1 first and then x1
#         arr = sorted(arr, key=lambda r: (r["page_number"], r["x0"], r["top"]))
#         for i in range(len(arr) - 1):
#             for j in range(i, -1, -1):
#                 # restore the order using th
#                 if abs(arr[j + 1]["x0"] - arr[j]["x0"]) < threashold \
#                         and arr[j + 1]["top"] < arr[j]["top"] \
#                         and arr[j + 1]["page_number"] == arr[j]["page_number"]:
#                     tmp = arr[j]
#                     arr[j] = arr[j + 1]
#                     arr[j + 1] = tmp
#         return arr
# 
#     def _has_color(self, o):
#         if o.get("ncs", "") == "DeviceGray":
#             if o["stroking_color"] and o["stroking_color"][0] == 1 and o["non_stroking_color"] and \
#                     o["non_stroking_color"][0] == 1:
#                 if re.match(r"[a-zT_\[\]\(\)-]+", o.get("text", "")):
#                     return False
#         return True
# 
#     def _table_transformer_job(self, ZM):
#         logging.info("Table processing...")
#         imgs, pos = [], []
#         tbcnt = [0]
#         MARGIN = 10
#         self.tb_cpns = []
#         assert len(self.page_layout) == len(self.page_images)
#         for p, tbls in enumerate(self.page_layout):  # for page
#             tbls = [f for f in tbls if f["type"] == "table"]
#             tbcnt.append(len(tbls))
#             if not tbls:
#                 continue
#             for tb in tbls:  # for table
#                 left, top, right, bott = tb["x0"] - MARGIN, tb["top"] - MARGIN, \
#                                          tb["x1"] + MARGIN, tb["bottom"] + MARGIN
#                 left *= ZM
#                 top *= ZM
#                 right *= ZM
#                 bott *= ZM
#                 pos.append((left, top))
#                 imgs.append(self.page_images[p].crop((left, top, right, bott)))
# 
#         assert len(self.page_images) == len(tbcnt) - 1
#         if not imgs:
#             return
#         recos = self.tbl_det(imgs)
#         tbcnt = np.cumsum(tbcnt)
#         for i in range(len(tbcnt) - 1):  # for page
#             pg = []
#             for j, tb_items in enumerate(
#                     recos[tbcnt[i]: tbcnt[i + 1]]):  # for table
#                 poss = pos[tbcnt[i]: tbcnt[i + 1]]
#                 for it in tb_items:  # for table components
#                     it["x0"] = (it["x0"] + poss[j][0])
#                     it["x1"] = (it["x1"] + poss[j][0])
#                     it["top"] = (it["top"] + poss[j][1])
#                     it["bottom"] = (it["bottom"] + poss[j][1])
#                     for n in ["x0", "x1", "top", "bottom"]:
#                         it[n] /= ZM
#                     it["top"] += self.page_cum_height[i]
#                     it["bottom"] += self.page_cum_height[i]
#                     it["pn"] = i
#                     it["layoutno"] = j
#                     pg.append(it)
#             self.tb_cpns.extend(pg)
# 
#         def gather(kwd, fzy=10, ption=0.6):
#             eles = Recognizer.sort_Y_firstly(
#                 [r for r in self.tb_cpns if re.match(kwd, r["label"])], fzy)
#             eles = Recognizer.layouts_cleanup(self.boxes, eles, 5, ption)
#             return Recognizer.sort_Y_firstly(eles, 0)
# 
#         # add R,H,C,SP tag to boxes within table layout
#         headers = gather(r".*header$")
#         rows = gather(r".* (row|header)")
#         spans = gather(r".*spanning")
#         clmns = sorted([r for r in self.tb_cpns if re.match(
#             r"table column$", r["label"])], key=lambda x: (x["pn"], x["layoutno"], x["x0"]))
#         clmns = Recognizer.layouts_cleanup(self.boxes, clmns, 5, 0.5)
#         for b in self.boxes:
#             if b.get("layout_type", "") != "table":
#                 continue
#             ii = Recognizer.find_overlapped_with_threashold(b, rows, thr=0.3)
#             if ii is not None:
#                 b["R"] = ii
#                 b["R_top"] = rows[ii]["top"]
#                 b["R_bott"] = rows[ii]["bottom"]
# 
#             ii = Recognizer.find_overlapped_with_threashold(
#                 b, headers, thr=0.3)
#             if ii is not None:
#                 b["H_top"] = headers[ii]["top"]
#                 b["H_bott"] = headers[ii]["bottom"]
#                 b["H_left"] = headers[ii]["x0"]
#                 b["H_right"] = headers[ii]["x1"]
#                 b["H"] = ii
# 
#             ii = Recognizer.find_horizontally_tightest_fit(b, clmns)
#             if ii is not None:
#                 b["C"] = ii
#                 b["C_left"] = clmns[ii]["x0"]
#                 b["C_right"] = clmns[ii]["x1"]
# 
#             ii = Recognizer.find_overlapped_with_threashold(b, spans, thr=0.3)
#             if ii is not None:
#                 b["H_top"] = spans[ii]["top"]
#                 b["H_bott"] = spans[ii]["bottom"]
#                 b["H_left"] = spans[ii]["x0"]
#                 b["H_right"] = spans[ii]["x1"]
#                 b["SP"] = ii
# 
#     def __ocr(self, pagenum, img, chars, ZM=3):
#         bxs = self.ocr.detect(np.array(img))
#         if not bxs:
#             self.boxes.append([])
#             return
#         bxs = [(line[0], line[1][0]) for line in bxs]
#         bxs = Recognizer.sort_Y_firstly(
#             [{"x0": b[0][0] / ZM, "x1": b[1][0] / ZM,
#               "top": b[0][1] / ZM, "text": "", "txt": t,
#               "bottom": b[-1][1] / ZM,
#               "page_number": pagenum} for b, t in bxs if b[0][0] <= b[1][0] and b[0][1] <= b[-1][1]],
#             self.mean_height[-1] / 3
#         )
# 
#         # merge chars in the same rect
#         for c in Recognizer.sort_X_firstly(
#                 chars, self.mean_width[pagenum - 1] // 4):
#             ii = Recognizer.find_overlapped(c, bxs)
#             if ii is None:
#                 self.lefted_chars.append(c)
#                 continue
#             ch = c["bottom"] - c["top"]
#             bh = bxs[ii]["bottom"] - bxs[ii]["top"]
#             if abs(ch - bh) / max(ch, bh) >= 0.7 and c["text"] != ' ':
#                 self.lefted_chars.append(c)
#                 continue
#             if c["text"] == " " and bxs[ii]["text"]:
#                 if re.match(r"[0-9a-zA-Z,.?;:!%%]", bxs[ii]["text"][-1]):
#                     bxs[ii]["text"] += " "
#             else:
#                 bxs[ii]["text"] += c["text"]
# 
#         for b in bxs:
#             if not b["text"]:
#                 left, right, top, bott = b["x0"] * ZM, b["x1"] * \
#                                          ZM, b["top"] * ZM, b["bottom"] * ZM
#                 b["text"] = self.ocr.recognize(np.array(img),
#                                                np.array([[left, top], [right, top], [right, bott], [left, bott]],
#                                                         dtype=np.float32))
#             del b["txt"]
#         bxs = [b for b in bxs if b["text"]]
#         if self.mean_height[-1] == 0:
#             self.mean_height[-1] = np.median([b["bottom"] - b["top"]
#                                               for b in bxs])
#         self.boxes.append(bxs)
# 
#     def _layouts_rec(self, ZM, drop=True):
#         assert len(self.page_images) == len(self.boxes)
#         self.boxes, self.page_layout = self.layouter(
#             self.page_images, self.boxes, ZM, drop=drop)
#         # cumlative Y
#         for i in range(len(self.boxes)):
#             self.boxes[i]["top"] += \
#                 self.page_cum_height[self.boxes[i]["page_number"] - 1]
#             self.boxes[i]["bottom"] += \
#                 self.page_cum_height[self.boxes[i]["page_number"] - 1]
# 
#     def _text_merge(self):
#         # merge adjusted boxes
#         bxs = self.boxes
# 
#         def end_with(b, txt):
#             txt = txt.strip()
#             tt = b.get("text", "").strip()
#             return tt and tt.find(txt) == len(tt) - len(txt)
# 
#         def start_with(b, txts):
#             tt = b.get("text", "").strip()
#             return tt and any([tt.find(t.strip()) == 0 for t in txts])
# 
#         # horizontally merge adjacent box with the same layout
#         i = 0
#         while i < len(bxs) - 1:
#             b = bxs[i]
#             b_ = bxs[i + 1]
#             if b.get("layoutno", "0") != b_.get("layoutno", "1") or b.get("layout_type", "") in ["table", "figure",
#                                                                                                  "equation"]:
#                 i += 1
#                 continue
#             if abs(self._y_dis(b, b_)
#                    ) < self.mean_height[bxs[i]["page_number"] - 1] / 3:
#                 # merge
#                 bxs[i]["x1"] = b_["x1"]
#                 bxs[i]["top"] = (b["top"] + b_["top"]) / 2
#                 bxs[i]["bottom"] = (b["bottom"] + b_["bottom"]) / 2
#                 bxs[i]["text"] += b_["text"]
#                 bxs.pop(i + 1)
#                 continue
#             i += 1
#             continue
# 
#             dis_thr = 1
#             dis = b["x1"] - b_["x0"]
#             if b.get("layout_type", "") != "text" or b_.get(
#                     "layout_type", "") != "text":
#                 if end_with(b, "，") or start_with(b_, "（，"):
#                     dis_thr = -8
#                 else:
#                     i += 1
#                     continue
# 
#             if abs(self._y_dis(b, b_)) < self.mean_height[bxs[i]["page_number"] - 1] / 5 \
#                     and dis >= dis_thr and b["x1"] < b_["x1"]:
#                 # merge
#                 bxs[i]["x1"] = b_["x1"]
#                 bxs[i]["top"] = (b["top"] + b_["top"]) / 2
#                 bxs[i]["bottom"] = (b["bottom"] + b_["bottom"]) / 2
#                 bxs[i]["text"] += b_["text"]
#                 bxs.pop(i + 1)
#                 continue
#             i += 1
#         self.boxes = bxs
# 
#     def _naive_vertical_merge(self):
#         bxs = Recognizer.sort_Y_firstly(
#             self.boxes, np.median(
#                 self.mean_height) / 3)
#         i = 0
#         while i + 1 < len(bxs):
#             b = bxs[i]
#             b_ = bxs[i + 1]
#             if b["page_number"] < b_["page_number"] and re.match(
#                     r"[0-9  •一—-]+$", b["text"]):
#                 bxs.pop(i)
#                 continue
#             if not b["text"].strip():
#                 bxs.pop(i)
#                 continue
#             concatting_feats = [
#                 b["text"].strip()[-1] in ",;:'\"，、‘“；：-",
#                 len(b["text"].strip()) > 1 and b["text"].strip(
#                 )[-2] in ",;:'\"，‘“、；：",
#                 b_["text"].strip() and b_["text"].strip()[0] in "。；？！?”）),，、：",
#             ]
#             # features for not concating
#             feats = [
#                 b.get("layoutno", 0) != b_.get("layoutno", 0),
#                 b["text"].strip()[-1] in "。？！?",
#                 self.is_english and b["text"].strip()[-1] in ".!?",
#                 b["page_number"] == b_["page_number"] and b_["top"] -
#                 b["bottom"] > self.mean_height[b["page_number"] - 1] * 1.5,
#                 b["page_number"] < b_["page_number"] and abs(
#                     b["x0"] - b_["x0"]) > self.mean_width[b["page_number"] - 1] * 4,
#             ]
#             # split features
#             detach_feats = [b["x1"] < b_["x0"],
#                             b["x0"] > b_["x1"]]
#             if (any(feats) and not any(concatting_feats)) or any(detach_feats):
#                 print(
#                     b["text"],
#                     b_["text"],
#                     any(feats),
#                     any(concatting_feats),
#                     any(detach_feats))
#                 i += 1
#                 continue
#             # merge up and down
#             b["bottom"] = b_["bottom"]
#             b["text"] += b_["text"]
#             b["x0"] = min(b["x0"], b_["x0"])
#             b["x1"] = max(b["x1"], b_["x1"])
#             bxs.pop(i + 1)
#         self.boxes = bxs
# 
#     def _concat_downward(self, concat_between_pages=True):
#         # count boxes in the same row as a feature
#         for i in range(len(self.boxes)):
#             mh = self.mean_height[self.boxes[i]["page_number"] - 1]
#             self.boxes[i]["in_row"] = 0
#             j = max(0, i - 12)
#             while j < min(i + 12, len(self.boxes)):
#                 if j == i:
#                     j += 1
#                     continue
#                 ydis = self._y_dis(self.boxes[i], self.boxes[j]) / mh
#                 if abs(ydis) < 1:
#                     self.boxes[i]["in_row"] += 1
#                 elif ydis > 0:
#                     break
#                 j += 1
# 
#         # concat between rows
#         boxes = deepcopy(self.boxes)
#         blocks = []
#         while boxes:
#             chunks = []
# 
#             def dfs(up, dp):
#                 chunks.append(up)
#                 i = dp
#                 while i < min(dp + 12, len(boxes)):
#                     ydis = self._y_dis(up, boxes[i])
#                     smpg = up["page_number"] == boxes[i]["page_number"]
#                     mh = self.mean_height[up["page_number"] - 1]
#                     mw = self.mean_width[up["page_number"] - 1]
#                     if smpg and ydis > mh * 4:
#                         break
#                     if not smpg and ydis > mh * 16:
#                         break
#                     down = boxes[i]
#                     if not concat_between_pages and down["page_number"] > up["page_number"]:
#                         break
# 
#                     if up.get("R", "") != down.get(
#                             "R", "") and up["text"][-1] != "，":
#                         i += 1
#                         continue
# 
#                     if re.match(r"[0-9]{2,3}/[0-9]{3}$", up["text"]) \
#                             or re.match(r"[0-9]{2,3}/[0-9]{3}$", down["text"]) \
#                             or not down["text"].strip():
#                         i += 1
#                         continue
# 
#                     if not down["text"].strip():
#                         i += 1
#                         continue
# 
#                     if up["x1"] < down["x0"] - 10 * \
#                             mw or up["x0"] > down["x1"] + 10 * mw:
#                         i += 1
#                         continue
# 
#                     if i - dp < 5 and up.get("layout_type") == "text":
#                         if up.get("layoutno", "1") == down.get(
#                                 "layoutno", "2"):
#                             dfs(down, i + 1)
#                             boxes.pop(i)
#                             return
#                         i += 1
#                         continue
# 
#                     fea = self._updown_concat_features(up, down)
#                     if self.updown_cnt_mdl.predict(
#                             xgb.DMatrix([fea]))[0] <= 0.5:
#                         i += 1
#                         continue
#                     dfs(down, i + 1)
#                     boxes.pop(i)
#                     return
# 
#             dfs(boxes[0], 1)
#             boxes.pop(0)
#             if chunks:
#                 blocks.append(chunks)
# 
#         # concat within each block
#         boxes = []
#         for b in blocks:
#             if len(b) == 1:
#                 boxes.append(b[0])
#                 continue
#             t = b[0]
#             for c in b[1:]:
#                 t["text"] = t["text"].strip()
#                 c["text"] = c["text"].strip()
#                 if not c["text"]:
#                     continue
#                 if t["text"] and re.match(
#                         r"[0-9\.a-zA-Z]+$", t["text"][-1] + c["text"][-1]):
#                     t["text"] += " "
#                 t["text"] += c["text"]
#                 t["x0"] = min(t["x0"], c["x0"])
#                 t["x1"] = max(t["x1"], c["x1"])
#                 t["page_number"] = min(t["page_number"], c["page_number"])
#                 t["bottom"] = c["bottom"]
#                 if not t["layout_type"] \
#                         and c["layout_type"]:
#                     t["layout_type"] = c["layout_type"]
#             boxes.append(t)
# 
#         self.boxes = Recognizer.sort_Y_firstly(boxes, 0)
# 
#     def _filter_forpages(self):
#         if not self.boxes:
#             return
#         findit = False
#         i = 0
#         while i < len(self.boxes):
#             if not re.match(r"(contents|目录|目次|table of contents|致谢|acknowledge)$",
#                             re.sub(r"( | |\u3000)+", "", self.boxes[i]["text"].lower())):
#                 i += 1
#                 continue
#             findit = True
#             eng = re.match(
#                 r"[0-9a-zA-Z :'.-]{5,}",
#                 self.boxes[i]["text"].strip())
#             self.boxes.pop(i)
#             if i >= len(self.boxes):
#                 break
#             prefix = self.boxes[i]["text"].strip()[:3] if not eng else " ".join(
#                 self.boxes[i]["text"].strip().split(" ")[:2])
#             while not prefix:
#                 self.boxes.pop(i)
#                 if i >= len(self.boxes):
#                     break
#                 prefix = self.boxes[i]["text"].strip()[:3] if not eng else " ".join(
#                     self.boxes[i]["text"].strip().split(" ")[:2])
#             self.boxes.pop(i)
#             if i >= len(self.boxes) or not prefix:
#                 break
#             for j in range(i, min(i + 128, len(self.boxes))):
#                 if not re.match(prefix, self.boxes[j]["text"]):
#                     continue
#                 for k in range(i, j):
#                     self.boxes.pop(i)
#                 break
#         if findit:
#             return
# 
#         page_dirty = [0] * len(self.page_images)
#         for b in self.boxes:
#             if re.search(r"(··|··|··)", b["text"]):
#                 page_dirty[b["page_number"] - 1] += 1
#         page_dirty = set([i + 1 for i, t in enumerate(page_dirty) if t > 3])
#         if not page_dirty:
#             return
#         i = 0
#         while i < len(self.boxes):
#             if self.boxes[i]["page_number"] in page_dirty:
#                 self.boxes.pop(i)
#                 continue
#             i += 1
# 
#     def _merge_with_same_bullet(self):
#         i = 0
#         while i + 1 < len(self.boxes):
#             b = self.boxes[i]
#             b_ = self.boxes[i + 1]
#             if not b["text"].strip():
#                 self.boxes.pop(i)
#                 continue
#             if not b_["text"].strip():
#                 self.boxes.pop(i + 1)
#                 continue
# 
#             if b["text"].strip()[0] != b_["text"].strip()[0] \
#                     or b["text"].strip()[0].lower() in set("qwertyuopasdfghjklzxcvbnm") \
#                     or rag_tokenizer.is_chinese(b["text"].strip()[0]) \
#                     or b["top"] > b_["bottom"]:
#                 i += 1
#                 continue
#             b_["text"] = b["text"] + "\n" + b_["text"]
#             b_["x0"] = min(b["x0"], b_["x0"])
#             b_["x1"] = max(b["x1"], b_["x1"])
#             b_["top"] = b["top"]
#             self.boxes.pop(i)
# 
#     def _extract_table_figure(self, need_image, ZM,
#                               return_html, need_position):
#         tables = {}
#         figures = {}
#         # extract figure and table boxes
#         i = 0
#         lst_lout_no = ""
#         nomerge_lout_no = []
#         while i < len(self.boxes):
#             if "layoutno" not in self.boxes[i]:
#                 i += 1
#                 continue
#             lout_no = str(self.boxes[i]["page_number"]) + \
#                       "-" + str(self.boxes[i]["layoutno"])
#             if TableStructureRecognizer.is_caption(self.boxes[i]) or self.boxes[i]["layout_type"] in ["table caption",
#                                                                                                       "title",
#                                                                                                       "figure caption",
#                                                                                                       "reference"]:
#                 nomerge_lout_no.append(lst_lout_no)
#             if self.boxes[i]["layout_type"] == "table":
#                 if re.match(r"(数据|资料|图表)*来源[:： ]", self.boxes[i]["text"]):
#                     self.boxes.pop(i)
#                     continue
#                 if lout_no not in tables:
#                     tables[lout_no] = []
#                 tables[lout_no].append(self.boxes[i])
#                 self.boxes.pop(i)
#                 lst_lout_no = lout_no
#                 continue
#             if need_image and self.boxes[i]["layout_type"] == "figure":
#                 if re.match(r"(数据|资料|图表)*来源[:： ]", self.boxes[i]["text"]):
#                     self.boxes.pop(i)
#                     continue
#                 if lout_no not in figures:
#                     figures[lout_no] = []
#                 figures[lout_no].append(self.boxes[i])
#                 self.boxes.pop(i)
#                 lst_lout_no = lout_no
#                 continue
#             i += 1
# 
#         # merge table on different pages
#         nomerge_lout_no = set(nomerge_lout_no)
#         tbls = sorted([(k, bxs) for k, bxs in tables.items()],
#                       key=lambda x: (x[1][0]["top"], x[1][0]["x0"]))
# 
#         i = len(tbls) - 1
#         while i - 1 >= 0:
#             k0, bxs0 = tbls[i - 1]
#             k, bxs = tbls[i]
#             i -= 1
#             if k0 in nomerge_lout_no:
#                 continue
#             if bxs[0]["page_number"] == bxs0[0]["page_number"]:
#                 continue
#             if bxs[0]["page_number"] - bxs0[0]["page_number"] > 1:
#                 continue
#             mh = self.mean_height[bxs[0]["page_number"] - 1]
#             if self._y_dis(bxs0[-1], bxs[0]) > mh * 23:
#                 continue
#             tables[k0].extend(tables[k])
#             del tables[k]
# 
#         def x_overlapped(a, b):
#             return not any([a["x1"] < b["x0"], a["x0"] > b["x1"]])
# 
#         # find captions and pop out
#         i = 0
#         while i < len(self.boxes):
#             c = self.boxes[i]
#             # mh = self.mean_height[c["page_number"]-1]
#             if not TableStructureRecognizer.is_caption(c):
#                 i += 1
#                 continue
# 
#             # find the nearest layouts
#             def nearest(tbls):
#                 nonlocal c
#                 mink = ""
#                 minv = 1000000000
#                 for k, bxs in tbls.items():
#                     for b in bxs:
#                         if b.get("layout_type", "").find("caption") >= 0:
#                             continue
#                         y_dis = self._y_dis(c, b)
#                         x_dis = self._x_dis(
#                             c, b) if not x_overlapped(
#                             c, b) else 0
#                         dis = y_dis * y_dis + x_dis * x_dis
#                         if dis < minv:
#                             mink = k
#                             minv = dis
#                 return mink, minv
# 
#             tk, tv = nearest(tables)
#             fk, fv = nearest(figures)
#             # if min(tv, fv) > 2000:
#             #    i += 1
#             #    continue
#             if tv < fv and tk:
#                 tables[tk].insert(0, c)
#                 logging.debug(
#                     "TABLE:" +
#                     self.boxes[i]["text"] +
#                     "; Cap: " +
#                     tk)
#             elif fk:
#                 figures[fk].insert(0, c)
#                 logging.debug(
#                     "FIGURE:" +
#                     self.boxes[i]["text"] +
#                     "; Cap: " +
#                     tk)
#             self.boxes.pop(i)
# 
#         res = []
#         positions = []
# 
#         def cropout(bxs, ltype, poss):
#             nonlocal ZM
#             pn = set([b["page_number"] - 1 for b in bxs])
#             if len(pn) < 2:
#                 pn = list(pn)[0]
#                 ht = self.page_cum_height[pn]
#                 b = {
#                     "x0": np.min([b["x0"] for b in bxs]),
#                     "top": np.min([b["top"] for b in bxs]) - ht,
#                     "x1": np.max([b["x1"] for b in bxs]),
#                     "bottom": np.max([b["bottom"] for b in bxs]) - ht
#                 }
#                 louts = [l for l in self.page_layout[pn] if l["type"] == ltype]
#                 ii = Recognizer.find_overlapped(b, louts, naive=True)
#                 if ii is not None:
#                     b = louts[ii]
#                 else:
#                     logging.warn(
#                         f"Missing layout match: {pn + 1},%s" %
#                         (bxs[0].get(
#                             "layoutno", "")))
# 
#                 left, top, right, bott = b["x0"], b["top"], b["x1"], b["bottom"]
#                 if right < left: right = left + 1
#                 poss.append((pn + self.page_from, left, right, top, bott))
#                 return self.page_images[pn] \
#                     .crop((left * ZM, top * ZM,
#                            right * ZM, bott * ZM))
#             pn = {}
#             for b in bxs:
#                 p = b["page_number"] - 1
#                 if p not in pn:
#                     pn[p] = []
#                 pn[p].append(b)
#             pn = sorted(pn.items(), key=lambda x: x[0])
#             imgs = [cropout(arr, ltype, poss) for p, arr in pn]
#             pic = Image.new("RGB",
#                             (int(np.max([i.size[0] for i in imgs])),
#                              int(np.sum([m.size[1] for m in imgs]))),
#                             (245, 245, 245))
#             height = 0
#             for img in imgs:
#                 pic.paste(img, (0, int(height)))
#                 height += img.size[1]
#             return pic
# 
#         # crop figure out and add caption
#         for k, bxs in figures.items():
#             txt = "\n".join([b["text"] for b in bxs])
#             if not txt:
#                 continue
# 
#             poss = []
#             res.append(
#                 (cropout(
#                     bxs,
#                     "figure", poss),
#                  [txt]))
#             positions.append(poss)
# 
#         for k, bxs in tables.items():
#             if not bxs:
#                 continue
#             bxs = Recognizer.sort_Y_firstly(bxs, np.mean(
#                 [(b["bottom"] - b["top"]) / 2 for b in bxs]))
#             poss = []
#             res.append((cropout(bxs, "table", poss),
#                         self.tbl_det.construct_table(bxs, html=return_html, is_english=self.is_english)))
#             positions.append(poss)
# 
#         assert len(positions) == len(res)
# 
#         if need_position:
#             return list(zip(res, positions))
#         return res
# 
#     def proj_match(self, line):
#         if len(line) <= 2:
#             return
#         if re.match(r"[0-9 ().,%%+/-]+$", line):
#             return False
#         for p, j in [
#             (r"第[零一二三四五六七八九十百]+章", 1),
#             (r"第[零一二三四五六七八九十百]+[条节]", 2),
#             (r"[零一二三四五六七八九十百]+[、 　]", 3),
#             (r"[\(（][零一二三四五六七八九十百]+[）\)]", 4),
#             (r"[0-9]+(、|\.[　 ]|\.[^0-9])", 5),
#             (r"[0-9]+\.[0-9]+(、|[. 　]|[^0-9])", 6),
#             (r"[0-9]+\.[0-9]+\.[0-9]+(、|[ 　]|[^0-9])", 7),
#             (r"[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+(、|[ 　]|[^0-9])", 8),
#             (r".{,48}[：:?？]$", 9),
#             (r"[0-9]+）", 10),
#             (r"[\(（][0-9]+[）\)]", 11),
#             (r"[零一二三四五六七八九十百]+是", 12),
#             (r"[⚫•➢✓]", 12)
#         ]:
#             if re.match(p, line):
#                 return j
#         return
# 
#     def _line_tag(self, bx, ZM):
#         pn = [bx["page_number"]]
#         top = bx["top"] - self.page_cum_height[pn[0] - 1]
#         bott = bx["bottom"] - self.page_cum_height[pn[0] - 1]
#         page_images_cnt = len(self.page_images)
#         if pn[-1] - 1 >= page_images_cnt: return ""
#         while bott * ZM > self.page_images[pn[-1] - 1].size[1]:
#             bott -= self.page_images[pn[-1] - 1].size[1] / ZM
#             pn.append(pn[-1] + 1)
#             if pn[-1] - 1 >= page_images_cnt:
#                 return ""
# 
#         return "@@{}\t{:.1f}\t{:.1f}\t{:.1f}\t{:.1f}##" \
#             .format("-".join([str(p) for p in pn]),
#                     bx["x0"], bx["x1"], top, bott)
# 
#     def __filterout_scraps(self, boxes, ZM):
# 
#         def width(b):
#             return b["x1"] - b["x0"]
# 
#         def height(b):
#             return b["bottom"] - b["top"]
# 
#         def usefull(b):
#             if b.get("layout_type"):
#                 return True
#             if width(
#                     b) > self.page_images[b["page_number"] - 1].size[0] / ZM / 3:
#                 return True
#             if b["bottom"] - b["top"] > self.mean_height[b["page_number"] - 1]:
#                 return True
#             return False
# 
#         res = []
#         while boxes:
#             lines = []
#             widths = []
#             pw = self.page_images[boxes[0]["page_number"] - 1].size[0] / ZM
#             mh = self.mean_height[boxes[0]["page_number"] - 1]
#             mj = self.proj_match(
#                 boxes[0]["text"]) or boxes[0].get(
#                 "layout_type",
#                 "") == "title"
# 
#             def dfs(line, st):
#                 nonlocal mh, pw, lines, widths
#                 lines.append(line)
#                 widths.append(width(line))
#                 width_mean = np.mean(widths)
#                 mmj = self.proj_match(
#                     line["text"]) or line.get(
#                     "layout_type",
#                     "") == "title"
#                 for i in range(st + 1, min(st + 20, len(boxes))):
#                     if (boxes[i]["page_number"] - line["page_number"]) > 0:
#                         break
#                     if not mmj and self._y_dis(
#                             line, boxes[i]) >= 3 * mh and height(line) < 1.5 * mh:
#                         break
# 
#                     if not usefull(boxes[i]):
#                         continue
#                     if mmj or \
#                             (self._x_dis(boxes[i], line) < pw / 10): \
#                             # and abs(width(boxes[i])-width_mean)/max(width(boxes[i]),width_mean)<0.5):
#                         # concat following
#                         dfs(boxes[i], i)
#                         boxes.pop(i)
#                         break
# 
#             try:
#                 if usefull(boxes[0]):
#                     dfs(boxes[0], 0)
#                 else:
#                     logging.debug("WASTE: " + boxes[0]["text"])
#             except Exception as e:
#                 pass
#             boxes.pop(0)
#             mw = np.mean(widths)
#             if mj or mw / pw >= 0.35 or mw > 200:
#                 res.append(
#                     "\n".join([c["text"] + self._line_tag(c, ZM) for c in lines]))
#             else:
#                 logging.debug("REMOVED: " +
#                               "<<".join([c["text"] for c in lines]))
# 
#         return "\n\n".join(res)
# 
#     @staticmethod
#     def total_page_number(fnm, binary=None):
#         try:
#             pdf = pdfplumber.open(
#                 fnm) if not binary else pdfplumber.open(BytesIO(binary))
#             return len(pdf.pages)
#         except Exception as e:
#             logging.error(str(e))
# 
#     def __images__(self, fnm, zoomin=3, page_from=0,
#                    page_to=299, callback=None):
#         self.lefted_chars = []
#         self.mean_height = []
#         self.mean_width = []
#         self.boxes = []
#         self.garbages = {}
#         self.page_cum_height = [0]
#         self.page_layout = []
#         self.page_from = page_from
#         st = timer()
#         try:
#             self.pdf = pdfplumber.open(fnm) if isinstance(
#                 fnm, str) else pdfplumber.open(BytesIO(fnm))
#             self.page_images = [p.to_image(resolution=72 * zoomin).annotated for i, p in
#                                 enumerate(self.pdf.pages[page_from:page_to])]
#             self.page_chars = [[c for c in page.chars if self._has_color(c)] for page in
#                                self.pdf.pages[page_from:page_to]]
#             self.total_page = len(self.pdf.pages)
#         except Exception as e:
#             logging.error(str(e))
# 
#         self.outlines = []
#         try:
#             self.pdf = pdf2_read(fnm if isinstance(fnm, str) else BytesIO(fnm))
#             outlines = self.pdf.outline
# 
#             def dfs(arr, depth):
#                 for a in arr:
#                     if isinstance(a, dict):
#                         self.outlines.append((a["/Title"], depth))
#                         continue
#                     dfs(a, depth + 1)
# 
#             dfs(outlines, 0)
#         except Exception as e:
#             logging.warning(f"Outlines exception: {e}")
#         if not self.outlines:
#             logging.warning(f"Miss outlines")
# 
#         logging.info("Images converted.")
#         self.is_english = [re.search(r"[a-zA-Z0-9,/¸;:'\[\]\(\)!@#$%^&*\"?<>._-]{30,}", "".join(
#             random.choices([c["text"] for c in self.page_chars[i]], k=min(100, len(self.page_chars[i]))))) for i in
#                            range(len(self.page_chars))]
#         if sum([1 if e else 0 for e in self.is_english]) > len(
#                 self.page_images) / 2:
#             self.is_english = True
#         else:
#             self.is_english = False
#         self.is_english = False
# 
#         st = timer()
#         for i, img in enumerate(self.page_images):
#             chars = self.page_chars[i] if not self.is_english else []
#             self.mean_height.append(
#                 np.median(sorted([c["height"] for c in chars])) if chars else 0
#             )
#             self.mean_width.append(
#                 np.median(sorted([c["width"] for c in chars])) if chars else 8
#             )
#             self.page_cum_height.append(img.size[1] / zoomin)
#             j = 0
#             while j + 1 < len(chars):
#                 if chars[j]["text"] and chars[j + 1]["text"] \
#                         and re.match(r"[0-9a-zA-Z,.:;!%]+", chars[j]["text"] + chars[j + 1]["text"]) \
#                         and chars[j + 1]["x0"] - chars[j]["x1"] >= min(chars[j + 1]["width"],
#                                                                        chars[j]["width"]) / 2:
#                     chars[j]["text"] += " "
#                 j += 1
# 
#             self.__ocr(i + 1, img, chars, zoomin)
#             if callback and i % 6 == 5:
#                 callback(prog=(i + 1) * 0.6 / len(self.page_images), msg="")
#         # print("OCR:", timer()-st)
# 
#         if not self.is_english and not any(
#                 [c for c in self.page_chars]) and self.boxes:
#             bxes = [b for bxs in self.boxes for b in bxs]
#             self.is_english = re.search(r"[\na-zA-Z0-9,/¸;:'\[\]\(\)!@#$%^&*\"?<>._-]{30,}",
#                                         "".join([b["text"] for b in random.choices(bxes, k=min(30, len(bxes)))]))
# 
#         logging.info("Is it English:", self.is_english)
# 
#         self.page_cum_height = np.cumsum(self.page_cum_height)
#         assert len(self.page_cum_height) == len(self.page_images) + 1
#         if len(self.boxes) == 0 and zoomin < 9: self.__images__(fnm, zoomin * 3, page_from,
#                                                                 page_to, callback)
# 
#     def __call__(self, fnm, need_image=True, zoomin=3, return_html=False):
#         self.__images__(fnm, zoomin)
#         self._layouts_rec(zoomin)
#         self._table_transformer_job(zoomin)
#         self._text_merge()
#         self._concat_downward()
#         self._filter_forpages()
#         tbls = self._extract_table_figure(
#             need_image, zoomin, return_html, False)
#         return self.__filterout_scraps(deepcopy(self.boxes), zoomin), tbls
# 
#     def remove_tag(self, txt):
#         return re.sub(r"@@[\t0-9.-]+?##", "", txt)
# 
#     def crop(self, text, ZM=3, need_position=False):
#         imgs = []
#         poss = []
#         for tag in re.findall(r"@@[0-9-]+\t[0-9.\t]+##", text):
#             pn, left, right, top, bottom = tag.strip(
#                 "#").strip("@").split("\t")
#             left, right, top, bottom = float(left), float(
#                 right), float(top), float(bottom)
#             poss.append(([int(p) - 1 for p in pn.split("-")],
#                          left, right, top, bottom))
#         if not poss:
#             if need_position:
#                 return None, None
#             return
# 
#         max_width = max(
#             np.max([right - left for (_, left, right, _, _) in poss]), 6)
#         GAP = 6
#         pos = poss[0]
#         poss.insert(0, ([pos[0][0]], pos[1], pos[2], max(
#             0, pos[3] - 120), max(pos[3] - GAP, 0)))
#         pos = poss[-1]
#         poss.append(([pos[0][-1]], pos[1], pos[2], min(self.page_images[pos[0][-1]].size[1] / ZM, pos[4] + GAP),
#                      min(self.page_images[pos[0][-1]].size[1] / ZM, pos[4] + 120)))
# 
#         positions = []
#         for ii, (pns, left, right, top, bottom) in enumerate(poss):
#             right = left + max_width
#             bottom *= ZM
#             for pn in pns[1:]:
#                 bottom += self.page_images[pn - 1].size[1]
#             imgs.append(
#                 self.page_images[pns[0]].crop((left * ZM, top * ZM,
#                                                right *
#                                                ZM, min(
#                     bottom, self.page_images[pns[0]].size[1])
#                                                ))
#             )
#             if 0 < ii < len(poss) - 1:
#                 positions.append((pns[0] + self.page_from, left, right, top, min(
#                     bottom, self.page_images[pns[0]].size[1]) / ZM))
#             bottom -= self.page_images[pns[0]].size[1]
#             for pn in pns[1:]:
#                 imgs.append(
#                     self.page_images[pn].crop((left * ZM, 0,
#                                                right * ZM,
#                                                min(bottom,
#                                                    self.page_images[pn].size[1])
#                                                ))
#                 )
#                 if 0 < ii < len(poss) - 1:
#                     positions.append((pn + self.page_from, left, right, 0, min(
#                         bottom, self.page_images[pn].size[1]) / ZM))
#                 bottom -= self.page_images[pn].size[1]
# 
#         if not imgs:
#             if need_position:
#                 return None, None
#             return
#         height = 0
#         for img in imgs:
#             height += img.size[1] + GAP
#         height = int(height)
#         width = int(np.max([i.size[0] for i in imgs]))
#         pic = Image.new("RGB",
#                         (width, height),
#                         (245, 245, 245))
#         height = 0
#         for ii, img in enumerate(imgs):
#             if ii == 0 or ii + 1 == len(imgs):
#                 img = img.convert('RGBA')
#                 overlay = Image.new('RGBA', img.size, (0, 0, 0, 0))
#                 overlay.putalpha(128)
#                 img = Image.alpha_composite(img, overlay).convert("RGB")
#             pic.paste(img, (0, int(height)))
#             height += img.size[1] + GAP
# 
#         if need_position:
#             return pic, positions
#         return pic
# 
#     def get_position(self, bx, ZM):
#         poss = []
#         pn = bx["page_number"]
#         top = bx["top"] - self.page_cum_height[pn - 1]
#         bott = bx["bottom"] - self.page_cum_height[pn - 1]
#         poss.append((pn, bx["x0"], bx["x1"], top, min(
#             bott, self.page_images[pn - 1].size[1] / ZM)))
#         while bott * ZM > self.page_images[pn - 1].size[1]:
#             bott -= self.page_images[pn - 1].size[1] / ZM
#             top = 0
#             pn += 1
#             poss.append((pn, bx["x0"], bx["x1"], top, min(
#                 bott, self.page_images[pn - 1].size[1] / ZM)))
#         return poss
# 
# 
# class PdfSimParser(object):
#     def parse(self, filename, from_page=0, to_page=100000, **kwargs):
#         self.outlines = []
#         lines = []
#         try:
#             self.pdf = pdf2_read(
#                 filename if isinstance(
#                     filename, str) else BytesIO(filename))
#             for page in self.pdf.pages[from_page:to_page]:
#                 lines.extend([t for t in page.extract_text().split("\n")])
# 
#             outlines = self.pdf.outline
# 
#             def dfs(arr, depth):
#                 for a in arr:
#                     if isinstance(a, dict):
#                         self.outlines.append((a["/Title"], depth))
#                         continue
#                     dfs(a, depth + 1)
# 
#             dfs(outlines, 0)
#         except Exception as e:
#             logging.warning(f"Outlines exception: {e}")
#         if not self.outlines:
#             logging.warning(f"Miss outlines")
# 
#         return lines
# 
#     def crop(self, ck, need_position):
#         raise NotImplementedError
# 
#     @staticmethod
#     def remove_tag(txt):
#         raise NotImplementedError
# 
# 
# if __name__ == "__main__":
#     pp=PdfSimParser()
#     contents=pp.parse('/data/users/searchgpt/yq/GoMate_dev/data/competitions/df/A_document/AZ06.pdf')
#     print(contents)


#!/usr/bin/env python
# -*- coding:utf-8 _*-
"""
@author:quincy qiang
@license: Apache Licence
@file: RagApplication.py
@time: 2024/05/20
@contact: yanqiangmiffy@gamil.com
@description:use openai etc. llm service demo
"""
import os
from trustrag.modules.document.common_parser import CommonParser
from trustrag.modules.generator.chat import DeepSeekChat
from trustrag.modules.reranker.bge_reranker import BgeReranker
from trustrag.modules.retrieval.dense_retriever import DenseRetriever
from trustrag.modules.document.chunk import TextChunker

class ApplicationConfig():
    def __init__(self):
        self.retriever_config = None
        self.rerank_config = None


class RagApplication():
    def __init__(self, config):
        self.config = config
        self.parser = CommonParser()
        self.retriever = DenseRetriever(self.config.retriever_config)
        self.reranker = BgeReranker(self.config.rerank_config)
        self.llm = DeepSeekChat(key=self.config.your_key)
        self.tc=TextChunker()
        self.rag_prompt="""请结合参考的上下文内容回答用户问题，如果上下文不能支撑用户问题，那么回答不知道或者我无法根据参考信息回答。
        问题: {question}
        可参考的上下文：
        ···
        {context}
        ···
        如果给定的上下文无法让你做出回答，请回答数据库中没有这个内容，你不知道。
        有用的回答:"""
    def init_vector_store(self):
        """

        """
        print("init_vector_store ... ")
        all_paragraphs = []
        all_chunks = []
        for filename in os.listdir(self.config.docs_path):
            file_path = os.path.join(self.config.docs_path, filename)
            try:
                paragraphs=self.parser.parse(file_path)
                all_paragraphs.append(paragraphs)
            except:
                pass
        print("chunking for paragraphs")
        for paragraphs in all_paragraphs:
            chunks=self.tc.get_chunks(paragraphs, 256)
            all_chunks.extend(chunks)
        self.retriever.build_from_texts(all_chunks)
        print("init_vector_store done! ")
        self.retriever.save_index(self.config.retriever_config.index_path)

    def load_vector_store(self):
        self.retriever.load_index(self.config.retriever_config.index_path)

    def add_document(self, file_path):
        chunks = self.parser.parse(file_path)
        for chunk in chunks:
            self.retriever.add_text(chunk)
        print("add_document done!")

    def chat(self, question: str = '', top_k: int = 5):
        contents = self.retriever.retrieve(query=question, top_k=top_k)
        contents = self.reranker.rerank(query=question, documents=[content['text'] for content in contents])
        content = '\n'.join([content['text'] for content in contents])

        system_prompt = "你是一个人工智能助手."
        prompt = self.rag_prompt.format(system_prompt=system_prompt, question=question,context=content)
        history = [
            {"role": "user", "content": prompt}
        ]
        gen_conf = {
            "temperature": 0.7,
            "max_tokens": 100
        }

        # 调用 chat 方法进行对话
        result = self.llm.chat(system=system_prompt, history=history, gen_conf=gen_conf)
        return result, history, contents



import base64
from typing import List, Dict, Any
from zhipuai import ZhipuAI
from PIL import Image
from trustrag.modules.retrieval.multimodal_retriever import MultimodalRetriever,MultimodalRetrieverConfig

class MultimodalRAG:
    def __init__(
            self,
            api_key: str,
            retriever_config: MultimodalRetrieverConfig,
            model_name: str = "glm-4v-plus",
            top_k: int = 3
    ):
        self.client = ZhipuAI(api_key=api_key)
        self.retriever = MultimodalRetriever(retriever_config)
        # self.retriever.load_index()
        self.model_name = model_name
        self.top_k = top_k

    def _prepare_context(self, results: List[Dict[str, Any]]) -> str:
        context = "基于以下相似图片信息：\n"
        for idx, result in enumerate(results, 1):
            context += f"{idx}. {result['text']} (相似度: {result['score']:.2f})\n"
        return context

    def _image_to_base64(self, image: Image) -> str:
        # Convert the image to RGB mode if it's in RGBA mode
        if image.mode == 'RGBA':
            image = image.convert('RGB')

        # Save the image to a BytesIO buffer in JPEG format
        buffered = io.BytesIO()
        image.save(buffered, format="JPEG")

        # Encode the image data to base64 and return it as a string
        return base64.b64encode(buffered.getvalue()).decode('utf-8')

    def chat(self, query: str, include_images: bool = True) -> str:
        # 1. 检索相似内容
        results = self.retriever.retrieve(query, top_k=self.top_k)

        # 2. 准备提示信息
        context = self._prepare_context(results)
        full_prompt = f"{context}\n用户问题: {query}\n请基于用户提供的图片和上述图片信息回答问题。"

        # 3. 准备消息内容
        messages = [{"role": "user", "content": []}]

        # 4. 如果需要，添加检索到的图片
        if include_images:
            for result in results:
                img_base64 = self._image_to_base64(result['image'])
                messages[0]["content"].append({
                    "type": "image_url",
                    "image_url": {"url": img_base64}
                })

        messages[0]["content"].append({"type": "text", "text": full_prompt})
        # 5. 调用API获取回答
        response = self.client.chat.completions.create(
            model=self.model_name,
            messages=messages
        )

        return results, response.choices[0].message.content

    def chat_with_image(self, query: str, image_path: str) -> str:
        # 1. 读取和编码用户提供的图片
        with open(image_path, 'rb') as img_file:
            user_img_base64 = base64.b64encode(img_file.read()).decode('utf-8')

        # 2. 检索相似内容
        results = self.retriever.retrieve(query, top_k=self.top_k)

        # 3. 准备提示信息
        context = self._prepare_context(results)
        full_prompt = f"{context}\n用户问题: {query}\n请基于用户提供的图片和上述相似图片信息回答问题。"

        # 4. 准备消息内容，首先添加用户的图片
        messages = [{
            "role": "user",
            "content": [
                {
                    "type": "image_url",
                    "image_url": {"url": user_img_base64}
                },
                {
                    "type": "text",
                    "text": full_prompt
                }
            ]
        }]

        # 5. 添加检索到的相似图片
        for result in results:
            img_base64 = self._image_to_base64(result['image'])
            messages[0]["content"].append({
                "type": "image_url",
                "image_url": {"url": img_base64}
            })

        # 6. 调用API获取回答
        response = self.client.chat.completions.create(
            model=self.model_name,
            messages=messages
        )

        return response.choices[0].message.content

if __name__ == '__main__':
    # 初始化配置
    car_retriever_config = MultimodalRetrieverConfig(
        model_name='ViT-B-16',
        index_path='./index_car',
        batch_size=32,
        dim=512,
        download_root="data/chinese-clip-vit-base-patch16/"
    )
    # 初始化
    car_rag = MultimodalRAG(
        api_key="xxx",
        retriever_config=car_retriever_config,
        top_k=1
    )
    query_text = "冷却系统检查"
    retrieve_results, response = car_rag.chat(query_text)
    retrieve_results, response


import re

import loguru

from trustrag.modules.document.chunk import TextChunker
from trustrag.modules.document.docx_parser import DocxParser
from trustrag.modules.document.excel_parser import ExcelParser
from trustrag.modules.document.html_parser import HtmlParser
from trustrag.modules.document.pdf_parser_fast import PdfSimParser
from trustrag.modules.document.ppt_parser import PptParser
from trustrag.modules.document.txt_parser import TextParser
from trustrag.modules.document.markdown_parser import MarkdownParser

class CommonParser():
    def __init__(self):
        self.tc = TextChunker()

    def parse(self, file_path):
        # 读取文件内容
        filename = file_path
        with open(file_path, 'rb') as f:
            content = f.read()
        # bytes_io = BytesIO(content)
        # 检测文件类型
        # mime = magic.Magic(mime=True)
        # file_type = mime.from_buffer(content)
        # loguru.logger.info(filename, file_type, mime)
        loguru.logger.info(filename)
        if re.search(r"\.docx$", filename, re.IGNORECASE):
            parser = DocxParser()
        elif re.search(r"\.pdf$", filename, re.IGNORECASE):
            parser = PdfSimParser()
        elif re.search(r"\.xlsx?$", filename, re.IGNORECASE):
            parser = ExcelParser()
        elif re.search(r"\.pptx$", filename, re.IGNORECASE):
            parser = PptParser()
        elif re.search(r"\.(txt|py|js|java|c|cpp|h|php|go|ts|sh|cs|kt)$", filename, re.IGNORECASE):
            parser = TextParser()
        elif re.search(r"\.(htm|html)$", filename, re.IGNORECASE):
            parser = HtmlParser()
        elif re.search(r"\.doc$", filename, re.IGNORECASE):
            parser = DocxParser()
        elif re.search(r"\.md$", filename, re.IGNORECASE):
            parser = MarkdownParser()
        else:
            raise NotImplementedError(
                "file type not supported yet(pdf, xlsx, doc, docx, txt supported)")
        contents = parser.parse(content)
        # loguru.logger.info(contents)
        # contents = self.tc.chunk_sentences(contents, chunk_size=512)
        return contents



#!/usr/bin/env python
# -*- coding:utf-8 _*-
"""
@author:quincy qiang
@license: Apache Licence
@file: RagApplication.py
@time: 2024/05/20
@contact: yanqiangmiffy@gamil.com
"""
import os

from trustrag.modules.citation.match_citation import MatchCitation
from trustrag.modules.document.common_parser import CommonParser
from trustrag.modules.generator.llm import GLM4Chat
from trustrag.modules.reranker.bge_reranker import BgeReranker
from trustrag.modules.retrieval.dense_retriever import DenseRetriever
from trustrag.modules.document.chunk import TextChunker

class ApplicationConfig():
    def __init__(self):
        self.retriever_config = None
        self.rerank_config = None


class RagApplication():
    def __init__(self, config):
        self.config = config
        self.parser = CommonParser()
        self.retriever = DenseRetriever(self.config.retriever_config)
        self.reranker = BgeReranker(self.config.rerank_config)
        self.llm = GLM4Chat(self.config.llm_model_path)
        self.mc = MatchCitation()
        self.tc=TextChunker()
    def init_vector_store(self):
        """

        """
        print("init_vector_store ... ")
        all_paragraphs = []
        all_chunks = []
        for filename in os.listdir(self.config.docs_path):
            file_path = os.path.join(self.config.docs_path, filename)
            try:
                paragraphs=self.parser.parse(file_path)
                all_paragraphs.append(paragraphs)
            except:
                pass
        print("chunking for paragraphs")
        for paragraphs in all_paragraphs:
            chunks=self.tc.chunk_sentences(paragraphs, 256)
            all_chunks.extend(chunks)
        self.retriever.build_from_texts(all_chunks)
        print("init_vector_store done! ")
        self.retriever.save_index(self.config.retriever_config.index_path)

    def load_vector_store(self):
        self.retriever.load_index(self.config.retriever_config.index_path)

    def add_document(self, file_path):
        chunks = self.parser.parse(file_path)
        for chunk in chunks:
            self.retriever.add_text(chunk)
        print("add_document done!")

    def chat(self, question: str = '', top_k: int = 5):
        contents = self.retriever.retrieve(query=question, top_k=top_k)
        contents = self.reranker.rerank(query=question, documents=[content['text'] for content in contents])
        content = '\n'.join([content['text'] for content in contents])
        print(contents)
        result, history = self.llm.chat(question, [], content)
        # result = self.mc.ground_response(
        #     response=response,
        #     evidences=[content['text'] for content in contents],
        #     selected_idx=[idx for idx in range(len(contents))],
        #     markdown=True
        # )
        return result, history, contents



from trustrag.modules.rewriter.hyde_rewriter import HydeRewriter


class RewriterApp():
    """改写模块，用于理解和优化用户的原始问题。

    实现包括
    1. HyDE。用生成式大模型预回答，返回生成的假定答案。
    2. ...

    """

    def __init__(self, component_name=None):
        """Init required rewriter according to component name."""
        self.rewriter_list = ['hyde']
        assert component_name in self.rewriter_list
        if component_name == 'hyde':
            self.rewriter = HyDE_rewriter()

    def run(self, query, temperature=1e-10):
        """Run the required rewriter"""
        if query is None:
            raise ValueError('原始问题不能为空')
        return self.rewriter.run(query, temperature)



import json

import chardet

from trustrag.modules.document.utils import PROJECT_BASE
from trustrag.modules.document.utils import find_codec


def get_encoding(file):
    with open(file, 'rb') as f:
        tmp = chardet.detect(f.read())
        return tmp['encoding']


class JsonParser(object):
    def parse(self, fnm, from_page=0, to_page=100000, **kwargs):
        if not isinstance(fnm, str):
            encoding = find_codec(fnm)
            txt = fnm.decode(encoding, errors="ignore")
        else:
            with open(fnm, "r", encoding=get_encoding(fnm)) as f:
                txt = f.read()
        # print(txt)
        data = json.loads(txt)
        # print(data)
        sections = []
        try:
            sections.append(
                {
                    'source': data['source'],
                    'title': data['title'],
                    'date': data['date'],
                    'sec_num': 0,
                    'content': data['title'] + '\n' + data['content'],
                }
            )
        except:
            if 'sections' in data:
                # for document in data['documents']:
                for section in data['sections']:
                    sections.append(
                        {
                            'source': data['file_name'],
                            'title': data['title'],
                            'date': data['date'],
                            'sec_num': section['sec_num'],
                            'content': section['sec_theme'] + '\n' + section['content'],
                            'chunks': [
                                section['sec_theme'] + '\n' +
                                chunk['content'] for chunk in section['chunks']
                            ]
                        }
                    )
        # print(len(sections),len(json_lines))
        return sections

    def crop(self, ck, need_position):
        raise NotImplementedError

    @staticmethod
    def remove_tag(txt):
        raise NotImplementedError


if __name__ == '__main__':
    jp = JsonParser()
    print(f'{PROJECT_BASE}/data/docs/final_data/《中办通报》.json')
    data = jp.parse(f'{PROJECT_BASE}/data/docs/final_data/《中办通报》.json')
    print(data[0])



from tqdm import tqdm
import chardet

def get_encoding(file):
    with open(file, 'rb') as f:
        tmp = chardet.detect(f.read())
        return tmp['encoding']

class TextParser(object):
    def parse(self, fnm, encoding="utf-8", from_page=0, to_page=100000, **kwargs):
        # 如果 fnm 不是字符串（假设是字节流等），则使用 find_codec 找到编码
        if not isinstance(fnm, str):
            encoding = get_encoding(fnm) if encoding is None else encoding
            txt = fnm.decode(encoding, errors="ignore")
        else:
            # 如果是字符串文件路径，且没有传入 encoding，则自动调用 get_encoding 检测编码
            encoding = get_encoding(fnm) if encoding is None else encoding
            with open(fnm, "r", encoding=encoding) as f:
                txt = f.read()

        sections = []
        # 按行进行解析并显示进度
        for sec in tqdm(txt.split("\n"), desc="Parsing"):
            if sec.strip():
                sections.append(sec)
        return sections

    def crop(self, ck, need_position):
        raise NotImplementedError

    @staticmethod
    def remove_tag(txt):
        raise NotImplementedError


#!/usr/bin/env python
# -*- coding:utf-8 _*-
"""
@author:quincy qiang
@license: Apache Licence
@file: parser.py
@time: 2024/05/24
@contact: yanqiangmiffy@gamil.com
"""
import typing
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Optional
from langchain.docstore.document import Document


class BaseParser(ABC):
    """
    BaseParser is an Abstract Base Class (ABC) that serves as a template for all parser objects.
    It contains the common attributes and methods that each parser should implement.
    """

    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    async def get_chunks(
        self,
        filepath: str,
        metadata: Optional[dict],
        *args,
        **kwargs,
    ) -> typing.List[Document]:
        """
        Abstract method. This should asynchronously read a file and return its content in chunks.

        Parameters:
            loaded_data_point (LoadedDataPoint): Loaded Document to read and parse.

        Returns:
            typing.List[Document]: A list of Document objects, each representing a chunk of the file.
        """
        pass


#!/usr/bin/env python
# -*- coding:utf-8 _*-
"""
@author: quincy qiang
@license: Apache Licence
@file: codeparser.py.py
@time: 2025/01/08
@contact: yanqiangmiffy@gmail.com
@software: PyCharm
@description: A custom Markdown parser for extracting and processing chunks from Markdown files.
"""
import re
from typing import List, Dict, Union

import chardet
from langchain_community.document_loaders import UnstructuredMarkdownLoader
from langchain_core.documents.base import Document


def get_encoding(file: Union[str, bytes]) -> str:
    """
    Detects the encoding of a given file.

    Args:
        file (Union[str, bytes]): The file path or byte stream to detect the encoding for.

    Returns:
        str: The detected encoding of the file.
    """
    with open(file, 'rb') as f:
        tmp = chardet.detect(f.read())
        return tmp['encoding']


class MarkdownParser:
    """
    Custom Markdown parser for extracting and processing chunks from Markdown files.
    """

    def parse(
        self,
        fnm: Union[str, bytes],
        encoding: str = "utf-8",
    ) -> List[str]:
        """
        Extracts chunks of content from a given Markdown file.

        Args:
            fnm (Union[str, bytes]): The file path or byte stream of the Markdown file.
            encoding (str, optional): The encoding to use when reading the file. Defaults to "utf-8".

        Returns:
            List[str]: A list of merged paragraphs extracted from the Markdown file.
        """
        # If fnm is not a string (assumed to be a byte stream), detect the encoding
        if not isinstance(fnm, str):
            encoding = get_encoding(fnm) if encoding is None else encoding
            content = fnm.decode(encoding, errors="ignore")
            documents = self.parse_markdown_to_documents(content)
        else:
            loader = UnstructuredMarkdownLoader(fnm, mode="elements")
            documents = loader.load()
        paragraphs = self.merge_header_contents(documents)
        return paragraphs

    def parse_markdown_to_documents(self, content: str) -> List[Document]:
        """
        Parses a Markdown string into a list of Document objects.

        Args:
            content (str): The Markdown content to parse.

        Returns:
            List[Document]: A list of Document objects representing the parsed Markdown content.
        """
        # Regular expression to match Markdown headings
        heading_pattern = re.compile(r'^(#+)\s*(.*)$', re.MULTILINE)

        # Store the parsed results
        documents = []

        # Split the content into sections
        sections = content.split('\n')

        for section in sections:
            # Check if the section is a heading
            heading_match = heading_pattern.match(section)
            if heading_match:
                # Calculate the depth of the heading
                current_depth = len(heading_match.group(1)) - 1
                # Extract the heading content
                page_content = heading_match.group(2).strip()
                # Add to the results
                documents.append(
                    Document(
                        page_content=page_content,
                        metadata={"category_depth": current_depth}
                    )
                )
            else:
                # If not a heading and the content is not empty, add to the results
                if section.strip():
                    documents.append(
                        Document(page_content=section.strip(), metadata={})
                    )
        return documents

    def merge_header_contents(self, documents: List[Document]) -> List[str]:
        """
        Merges headers and their corresponding content into a list of paragraphs.

        Args:
            documents (List[Document]): A list of Document objects representing the parsed Markdown content.

        Returns:
            List[str]: A list of merged paragraphs, each containing a header and its corresponding content.
        """
        merged_data = []
        current_title = None
        current_content = []

        for document in documents:
            metadata = document.metadata
            category_depth = metadata.get('category_depth', None)
            page_content = document.page_content

            # If category_depth is 0, it indicates a top-level heading
            if category_depth == 0:
                # If current_title is not None, it means we have collected a complete heading and its content
                if current_title is not None:
                    # Merge the current title and content into a single string and add to merged_data
                    merged_content = "\n".join(current_content)
                    merged_data.append({
                        'title': current_title,
                        'content': merged_content
                    })
                    # Reset the current title and content
                    current_content = []

                # Update the current title and add Markdown heading markers based on category_depth
                current_title = f"{'#' * (category_depth + 1)} {page_content}"

            # If category_depth is not 0, it indicates body content or other headings
            else:
                # If current_title is None, it means the content starts with body text
                if current_title is None:
                    merged_data.append({
                        'title': '',
                        'content': page_content
                    })
                # Headings other than top-level (e.g., second-level, third-level, etc.)
                elif category_depth is not None:
                    # Add Markdown heading markers
                    current_content.append(f"{'#' * (category_depth + 1)} {document.page_content}")
                else:
                    # Add the content to the current content list
                    current_content.append(page_content)

        # Handle the last heading and its content
        if current_title is not None:
            merged_content = "\n".join(current_content)
            merged_data.append({
                'title': current_title,
                'content': merged_content
            })
        paragraphs = [item["title"] + "\n" + item["content"] for item in merged_data]

        return paragraphs


from trustrag.modules.reranker.bge_reranker import BgeReranker,BgeRerankerConfig


class RerankerApp():
    """重排模块，评估文档的相关性并重新排序。把最有可能提供准确、相关回答的文档排在前面。

    实现包括
    1. bge-reranker-large。智源开源的Rerank模型。
    2. ...

    """

    def __init__(self, component_name=None):
        """Init required reranker according to component name."""
        self.reranker_list = ['bge_large']
        assert component_name in self.reranker_list
        if component_name == 'bge_large':
            reranker_config = BgeRerankerConfig(
                model_name_or_path="/home/test/pretrained_models/bge-reranker-large"
            )
            self.reranker = BgeReranker(reranker_config)

    def run(self, query, contexts):
        """Run the required reranker"""
        if query is None:
            raise ValueError('missing query')
        if contexts is None:
            raise ValueError('missing contexts')
        return self.reranker.rerank(query, contexts)



# -*- coding: utf-8 -*-
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import chardet
import html_text
import readability

from trustrag.modules.document.utils import find_codec


def get_encoding(file):
    with open(file, 'rb') as f:
        tmp = chardet.detect(f.read())
        return tmp['encoding']


class HtmlParser:
    def parse(self, fnm):
        txt = ""
        if not isinstance(
            fnm, str):
            encoding = find_codec(fnm)
            txt = fnm.decode(encoding, errors="ignore")
        else:
            with open(fnm, "r", encoding=get_encoding(fnm)) as f:
                txt = f.read()

        html_doc = readability.Document(txt)
        title = html_doc.title()
        content = html_text.extract_text(html_doc.summary(html_partial=True))
        txt = f'{title}\n{content}'
        sections = txt.split("\n")
        return sections


if __name__ == '__main__':
    hp = HtmlParser()
    contents = hp.parse('/data/users/searchgpt/yq/GoMate_dev/data/docs/如何打通发展新质生产力的堵点卡点_新浪新闻.html')
    print(contents)



#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from io import BytesIO
from pptx import Presentation


class PptParser(object):
    def __init__(self):
        super().__init__()

    def __extract(self, shape):
        if shape.shape_type == 19:
            tb = shape.table
            rows = []
            for i in range(1, len(tb.rows)):
                rows.append("; ".join([tb.cell(
                    0, j).text + ": " + tb.cell(i, j).text for j in range(len(tb.columns)) if tb.cell(i, j)]))
            return "\n".join(rows)

        if shape.has_text_frame:
            return shape.text_frame.text

        if shape.shape_type == 6:
            texts = []
            for p in sorted(shape.shapes, key=lambda x: (x.top // 10, x.left)):
                t = self.__extract(p)
                if t:
                    texts.append(t)
            return "\n".join(texts)

    def parse(self, fnm, from_page=0, to_page=100000, callback=None):
        ppt = Presentation(fnm) if isinstance(
            fnm, str) else Presentation(
            BytesIO(fnm))
        txts = []
        self.total_page = len(ppt.slides)
        for i, slide in enumerate(ppt.slides):
            if i < from_page:
                continue
            if i >= to_page:
                break
            texts = []
            for shape in sorted(
                    slide.shapes, key=lambda x: (x.top // 10, x.left)):
                txt = self.__extract(shape)
                if txt:
                    texts.append(txt)
            txts.append("\n".join(texts))

        return txts

if __name__ == '__main__':
    pp=PptParser()
    txts=pp.parse('/data/users/searchgpt/yq/GoMate/data/docs/集成电路封装材料-光敏材料.pptx')
    print(txts)



#!/usr/bin/env python
# -*- coding:utf-8 _*-
"""
@author:quincy qiang
@license: Apache Licence
@file: pdfparser_fast.py
@time: 2024/06/06
@contact: yanqiangmiffy@gamil.com
@software: PyCharm
@description: coding..
"""
import re
from io import BytesIO
from PyPDF2 import PdfReader as pdf2_read
import fitz
from tqdm import tqdm
import logging

class PdfParserUsingPyMuPDF():
    """
    PdfParserUsingPyMuPDF is a parser class for extracting text from PDF files using PyMuPDF library.
    """

    supported_file_extensions = [".pdf"]

    def __init__(self, max_chunk_size: int = 1000, *args, **kwargs):
        """
        Initializes the PdfParserUsingPyMuPDF object.
        """
        self.max_chunk_size = max_chunk_size

    def parse(
            self, fnm: str, *args, **kwargs
    ):
        """
        Asynchronously extracts text from a PDF file and returns it in chunks.
        """
        final_texts = []
        final_tables = []
        # Open the PDF file using pdfplumber
        doc = fitz.open(fnm) if isinstance(
            fnm, str) else fitz.open(
            BytesIO(fnm))
        for page in tqdm(doc, total=len(doc), desc="get pages"):
            table = page.find_tables()
            table = list(table)
            for ix, tab in enumerate(table):
                tab = tab.extract()
                tab = list(map(lambda x: [str(t) for t in x], tab))
                tab = list(map("||".join, tab))
                tab = "\n".join(tab)
                final_tables.append(tab)

            text = page.get_text()
            # print(text)
            # clean up text for any problematic characters
            text = re.sub("\n", " ", text).strip()
            # text = text.encode("ascii", errors="ignore").decode("ascii")
            # text = re.sub(r"([^\w\s])\1{4,}", " ", text)
            # text = re.sub(" +", " ", text).strip()

            final_texts.append(text)

        return final_texts + final_tables

class PdfSimParser(object):
    def parse(self, filename, from_page=0, to_page=100000, **kwargs):
        self.outlines = []
        lines = []
        try:
            self.pdf = pdf2_read(
                filename if isinstance(
                    filename, str) else BytesIO(filename))
            for page in self.pdf.pages[from_page:to_page]:
                lines.extend([t for t in page.extract_text().split("\n")])

            outlines = self.pdf.outline

            def dfs(arr, depth):
                for a in arr:
                    if isinstance(a, dict):
                        self.outlines.append((a["/Title"], depth))
                        continue
                    dfs(a, depth + 1)

            dfs(outlines, 0)
        except Exception as e:
            logging.warning(f"Outlines exception: {e}")
        if not self.outlines:
            logging.warning(f"Miss outlines")

        return lines

    def crop(self, ck, need_position):
        raise NotImplementedError

    @staticmethod
    def remove_tag(txt):
        raise NotImplementedError

if __name__ == '__main__':
    pdf_parser = PdfParserUsingPyMuPDF()
    contents = pdf_parser.parse('/data/users/searchgpt/yq/GoMate_dev/data/docs/新冠肺炎疫情.pdf')
    print(contents)



#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from io import BytesIO
from openpyxl import load_workbook
from trustrag.modules.document.utils import find_codec


class ExcelParser:
    def html(self, fnm, chunk_rows=256):
        if isinstance(fnm, str):
            wb = load_workbook(fnm)
        else:
            wb = load_workbook(BytesIO(fnm))

        tb_chunks = []
        for sheetname in wb.sheetnames:
            ws = wb[sheetname]
            rows = list(ws.rows)
            if not rows: continue

            tb_rows_0 = "<tr>"
            for t in list(rows[0]):
                tb_rows_0 += f"<th>{t.value}</th>"
            tb_rows_0 += "</tr>"

            for chunk_i in range((len(rows) - 1) // chunk_rows + 1):
                tb = ""
                tb += f"<table><caption>{sheetname}</caption>"
                tb += tb_rows_0
                for r in list(rows[1 + chunk_i * chunk_rows:1 + (chunk_i + 1) * chunk_rows]):
                    tb += "<tr>"
                    for i, c in enumerate(r):
                        if c.value is None:
                            tb += "<td></td>"
                        else:
                            tb += f"<td>{c.value}</td>"
                    tb += "</tr>"
                tb += "</table>\n"
                tb_chunks.append(tb)

        return tb_chunks

    def parse(self, fnm):
        if isinstance(fnm, str):
            wb = load_workbook(fnm)
        else:
            wb = load_workbook(BytesIO(fnm))
        res = []
        for sheetname in wb.sheetnames:
            ws = wb[sheetname]
            rows = list(ws.rows)
            if not rows: continue
            ti = list(rows[0])
            for r in list(rows[1:]):
                l = []
                for i, c in enumerate(r):
                    if not c.value:
                        continue
                    t = str(ti[i].value) if i < len(ti) else ""
                    t += ("：" if t else "") + str(c.value)
                    l.append(t)
                l = "; ".join(l)
                if sheetname.lower().find("sheet") < 0:
                    l += " ——" + sheetname
                res.append(l)
        return res

    @staticmethod
    def row_number(fnm, binary):
        if fnm.split(".")[-1].lower().find("xls") >= 0:
            wb = load_workbook(BytesIO(binary))
            total = 0
            for sheetname in wb.sheetnames:
                ws = wb[sheetname]
                total += len(list(ws.rows))
                return total

        if fnm.split(".")[-1].lower() in ["csv", "txt"]:
            encoding = find_codec(binary)
            txt = binary.decode(encoding, errors="ignore")
            return len(txt.split("\n"))


if __name__ == "__main__":
    psr = ExcelParser()
    contents=psr.parse('/data/users/searchgpt/yq/GoMate_dev/data/docs/dev.xlsx')
    print(len(contents))
    print(contents[0])


import re
from tqdm import tqdm
from trustrag.modules.document import rag_tokenizer


class TextChunker:
    def __init__(self):
        """
        Initializes the TextChunker with a tokenizer.
        """
        self.tokenizer = rag_tokenizer

    def split_sentences(self, text: str) -> list[str]:
        """
        Splits the input text into sentences based on Chinese and English punctuation marks.

        Args:
            text (str): The input text to be split into sentences.

        Returns:
            list[str]: A list of sentences extracted from the input text.
        """
        # Use regex to split text by sentence-ending punctuation marks
        sentence_endings = re.compile(r'([。！？.!?])')
        sentences = sentence_endings.split(text)

        # Merge punctuation marks with their preceding sentences
        result = []
        for i in range(0, len(sentences) - 1, 2):
            if sentences[i]:
                result.append(sentences[i] + sentences[i + 1])

        # Handle the last sentence if it lacks punctuation
        if sentences[-1]:
            result.append(sentences[-1])

        # Remove whitespace and filter out empty sentences
        result = [sentence.strip() for sentence in result if sentence.strip()]

        return result

    def process_text_chunks(self, chunks: list[str]) -> list[str]:
        """
        Preprocesses text chunks by normalizing excessive newlines and spaces.

        Args:
            chunks (list[str]): A list of text chunks to be processed.

        Returns:
            list[str]: A list of processed text chunks with normalized formatting.
        """
        processed_chunks = []
        for chunk in chunks:
            # Normalize four or more consecutive newlines
            while '\n\n\n\n' in chunk:
                chunk = chunk.replace('\n\n\n\n', '\n\n')

            # Normalize four or more consecutive spaces
            while '    ' in chunk:
                chunk = chunk.replace('    ', '  ')

            processed_chunks.append(chunk)

        return processed_chunks

    def get_chunks(self, paragraphs: list[str], chunk_size: int) -> list[str]:
        """
        Splits a list of paragraphs into chunks based on a specified token size.

        Args:
            paragraphs (list[str]): A list of paragraphs to be chunked.
            chunk_size (int): The maximum number of tokens allowed per chunk.

        Returns:
            list[str]: A list of text chunks, each containing sentences that fit within the token limit.
        """
        # Combine paragraphs into a single text
        text = ''.join(paragraphs)

        # Split the text into sentences
        sentences = self.split_sentences(text)

        # If no sentences are found, treat paragraphs as sentences
        if len(sentences) == 0:
            sentences = paragraphs

        chunks = []
        current_chunk = []
        current_chunk_tokens = 0

        # Iterate through sentences and build chunks based on token count
        for sentence in sentences:
            tokens = self.tokenizer.tokenize(sentence)
            if current_chunk_tokens + len(tokens) <= chunk_size:
                # Add sentence to the current chunk if it fits
                current_chunk.append(sentence)
                current_chunk_tokens += len(tokens)
            else:
                # Finalize the current chunk and start a new one
                chunks.append(''.join(current_chunk))
                current_chunk = [sentence]
                current_chunk_tokens = len(tokens)

        # Add the last chunk if it contains any sentences
        if current_chunk:
            chunks.append(''.join(current_chunk))

        # Preprocess the chunks to normalize formatting
        chunks = self.process_text_chunks(chunks)
        return chunks


if __name__ == '__main__':
    # Example usage
    paragraphs = ['Hello!\nHi!\nGoodbye!']
    tc = TextChunker()
    chunk_size = 512
    chunks = tc.get_chunks(paragraphs, chunk_size)

    # Print the resulting chunks
    for i, chunk in enumerate(chunks):
        print(f"Chunk {i + 1}: {chunk}\n")


from setuptools import setup, find_packages
try:
    from trustrag.version import __version__
except ImportError:
    __version__ = "unknown version"

# 读取 requirements.txt 文件中的内容
with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name="trustrag",
    version=__version__,
    author="gomate-community",
    packages=['trustrag'],
    package_dir={'trustrag': 'trustrag'},
    package_data={'trustrag': ['*.*', 'applications/*', 'modules/document/*']},
    install_requires=required,
    author_email="yanqiang@ict.ac.cn",
    description="RAG Framework within Reliable input,Trusted output",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/gomate-community/TrustRAG",
    python_requires='>=3.9',
)


#!/usr/bin/env python
# -*- coding:utf-8 _*-
"""
@author:quincy qiang
@license: Apache Licence
@file: rag_tokenizer.py
@time: 2024/06/06
@contact: yanqiangmiffy@gamil.com
@software: PyCharm
@description: coding..
"""
import copy
import math
import os
import re
import string
import sys

import datrie
from hanziconv import HanziConv
from nltk import word_tokenize
from nltk.stem import PorterStemmer, WordNetLemmatizer

# from trustrag.modules.document.utils import get_project_base_directory

_get_module_path = lambda path: os.path.normpath(os.path.join(os.getcwd(),
                                                              os.path.dirname(__file__), path))
DEFAULT_IDF = _get_module_path("huqie.txt")
DEFAULT_IDF_TRIE = _get_module_path("huqie.txt.trie")

print("RAG Vocab:", DEFAULT_IDF)
print("RAG Trie:", DEFAULT_IDF_TRIE)


class RagTokenizer:
    def key_(self, line):
        return str(line.lower().encode("utf-8", 'ignore'))[2:-1]

    def rkey_(self, line):
        return str(("DD" + (line[::-1].lower())).encode("utf-8", 'ignore'))[2:-1]

    def loadDict_(self, fnm):
        print("[HUQIE]:Build trie", fnm, file=sys.stderr)
        try:
            of = open(fnm, "r", encoding='utf-8')
            while True:
                line = of.readline()
                if not line:
                    break
                line = re.sub(r"[\r\n]+", "", line)
                line = re.split(r"[ \t]", line)
                k = self.key_(line[0])
                F = int(math.log(float(line[1]) / self.DENOMINATOR) + .5)
                if k not in self.trie_ or self.trie_[k][0] < F:
                    self.trie_[self.key_(line[0])] = (F, line[2])
                self.trie_[self.rkey_(line[0])] = 1
            self.trie_.save(fnm + ".trie")
            of.close()
        except Exception as e:
            print("[HUQIE]:Faild to build trie, ", fnm, e, file=sys.stderr)

    def __init__(self, debug=False):
        self.DEBUG = debug
        self.DENOMINATOR = 1000000
        self.trie_ = datrie.Trie(string.printable)
        # self.DIR_ = os.path.join(get_project_base_directory(), "rag/res", "huqie")
        # self.DIR_ = os.path.join(get_project_base_directory(), "data/huqie", "huqie")

        self.stemmer = PorterStemmer()
        self.lemmatizer = WordNetLemmatizer()

        self.SPLIT_CHAR = r"([ ,\.<>/?;'\[\]\\`!@#$%^&*\(\)\{\}\|_+=《》，。？、；‘’：“”【】~！￥%……（）——-]+|[a-z\.-]+|[0-9,\.-]+)"
        try:
            self.trie_ = datrie.Trie.load(DEFAULT_IDF_TRIE)
            return
        except Exception as e:
            print("[HUQIE]:Build default trie", file=sys.stderr)
            self.trie_ = datrie.Trie(string.printable)

        self.loadDict_(DEFAULT_IDF)

    def loadUserDict(self, fnm):
        try:
            self.trie_ = datrie.Trie.load(DEFAULT_IDF_TRIE)
            return
        except Exception as e:
            self.trie_ = datrie.Trie(string.printable)
        self.loadDict_(fnm)

    def addUserDict(self, fnm):
        self.loadDict_(fnm)

    def _strQ2B(self, ustring):
        """把字符串全角转半角"""
        rstring = ""
        for uchar in ustring:
            inside_code = ord(uchar)
            if inside_code == 0x3000:
                inside_code = 0x0020
            else:
                inside_code -= 0xfee0
            if inside_code < 0x0020 or inside_code > 0x7e:  # 转完之后不是半角字符返回原来的字符
                rstring += uchar
            else:
                rstring += chr(inside_code)
        return rstring

    def _tradi2simp(self, line):
        return HanziConv.toSimplified(line)

    def dfs_(self, chars, s, preTks, tkslist):
        MAX_L = 10
        res = s
        # if s > MAX_L or s>= len(chars):
        if s >= len(chars):
            tkslist.append(preTks)
            return res

        # pruning
        S = s + 1
        if s + 2 <= len(chars):
            t1, t2 = "".join(chars[s:s + 1]), "".join(chars[s:s + 2])
            if self.trie_.has_keys_with_prefix(self.key_(t1)) and not self.trie_.has_keys_with_prefix(
                    self.key_(t2)):
                S = s + 2
        if len(preTks) > 2 and len(
                preTks[-1][0]) == 1 and len(preTks[-2][0]) == 1 and len(preTks[-3][0]) == 1:
            t1 = preTks[-1][0] + "".join(chars[s:s + 1])
            if self.trie_.has_keys_with_prefix(self.key_(t1)):
                S = s + 2

        ################
        for e in range(S, len(chars) + 1):
            t = "".join(chars[s:e])
            k = self.key_(t)

            if e > s + 1 and not self.trie_.has_keys_with_prefix(k):
                break

            if k in self.trie_:
                pretks = copy.deepcopy(preTks)
                if k in self.trie_:
                    pretks.append((t, self.trie_[k]))
                else:
                    pretks.append((t, (-12, '')))
                res = max(res, self.dfs_(chars, e, pretks, tkslist))

        if res > s:
            return res

        t = "".join(chars[s:s + 1])
        k = self.key_(t)
        if k in self.trie_:
            preTks.append((t, self.trie_[k]))
        else:
            preTks.append((t, (-12, '')))

        return self.dfs_(chars, s + 1, preTks, tkslist)

    def freq(self, tk):
        k = self.key_(tk)
        if k not in self.trie_:
            return 0
        return int(math.exp(self.trie_[k][0]) * self.DENOMINATOR + 0.5)

    def tag(self, tk):
        k = self.key_(tk)
        if k not in self.trie_:
            return ""
        return self.trie_[k][1]

    def score_(self, tfts):
        B = 30
        F, L, tks = 0, 0, []
        for tk, (freq, tag) in tfts:
            F += freq
            L += 0 if len(tk) < 2 else 1
            tks.append(tk)
        F /= len(tks)
        L /= len(tks)
        if self.DEBUG:
            print("[SC]", tks, len(tks), L, F, B / len(tks) + L + F)
        return tks, B / len(tks) + L + F

    def sortTks_(self, tkslist):
        res = []
        for tfts in tkslist:
            tks, s = self.score_(tfts)
            res.append((tks, s))
        return sorted(res, key=lambda x: x[1], reverse=True)

    def merge_(self, tks):
        patts = [
            (r"[ ]+", " "),
            (r"([0-9\+\.,%\*=-]) ([0-9\+\.,%\*=-])", r"\1\2"),
        ]
        # for p,s in patts: tks = re.sub(p, s, tks)

        # if split chars is part of token
        res = []
        tks = re.sub(r"[ ]+", " ", tks).split(" ")
        s = 0
        while True:
            if s >= len(tks):
                break
            E = s + 1
            for e in range(s + 2, min(len(tks) + 2, s + 6)):
                tk = "".join(tks[s:e])
                if re.search(self.SPLIT_CHAR, tk) and self.freq(tk):
                    E = e
            res.append("".join(tks[s:E]))
            s = E

        return res

    def maxForward_(self, line):
        res = []
        s = 0
        while s < len(line):
            e = s + 1
            t = line[s:e]
            while e < len(line) and self.trie_.has_keys_with_prefix(
                    self.key_(t)):
                e += 1
                t = line[s:e]

            while e - 1 > s and self.key_(t) not in self.trie_:
                e -= 1
                t = line[s:e]

            if self.key_(t) in self.trie_:
                res.append((t, self.trie_[self.key_(t)]))
            else:
                res.append((t, (0, '')))

            s = e

        return self.score_(res)

    def maxBackward_(self, line):
        res = []
        s = len(line) - 1
        while s >= 0:
            e = s + 1
            t = line[s:e]
            while s > 0 and self.trie_.has_keys_with_prefix(self.rkey_(t)):
                s -= 1
                t = line[s:e]

            while s + 1 < e and self.key_(t) not in self.trie_:
                s += 1
                t = line[s:e]

            if self.key_(t) in self.trie_:
                res.append((t, self.trie_[self.key_(t)]))
            else:
                res.append((t, (0, '')))

            s -= 1

        return self.score_(res[::-1])

    def english_normalize_(self, tks):
        return [self.stemmer.stem(self.lemmatizer.lemmatize(t)) if re.match(r"[a-zA-Z_-]+$", t) else t for t in tks]

    def tokenize(self, line):
        line = self._strQ2B(line).lower()
        line = self._tradi2simp(line)
        zh_num = len([1 for c in line if is_chinese(c)])
        if zh_num == 0:
            return " ".join([self.stemmer.stem(self.lemmatizer.lemmatize(t)) for t in word_tokenize(line)])

        arr = re.split(self.SPLIT_CHAR, line)
        res = []
        for L in arr:
            if len(L) < 2 or re.match(
                    r"[a-z\.-]+$", L) or re.match(r"[0-9\.-]+$", L):
                res.append(L)
                continue
            # print(L)

            # use maxforward for the first time
            tks, s = self.maxForward_(L)
            tks1, s1 = self.maxBackward_(L)
            if self.DEBUG:
                print("[FW]", tks, s)
                print("[BW]", tks1, s1)

            diff = [0 for _ in range(max(len(tks1), len(tks)))]
            for i in range(min(len(tks1), len(tks))):
                if tks[i] != tks1[i]:
                    diff[i] = 1

            if s1 > s:
                tks = tks1

            i = 0
            while i < len(tks):
                s = i
                while s < len(tks) and diff[s] == 0:
                    s += 1
                if s == len(tks):
                    res.append(" ".join(tks[i:]))
                    break
                if s > i:
                    res.append(" ".join(tks[i:s]))

                e = s
                while e < len(tks) and e - s < 5 and diff[e] == 1:
                    e += 1

                tkslist = []
                self.dfs_("".join(tks[s:e + 1]), 0, [], tkslist)
                res.append(" ".join(self.sortTks_(tkslist)[0][0]))

                i = e + 1

        res = " ".join(self.english_normalize_(res))
        if self.DEBUG:
            print("[TKS]", self.merge_(res))
        return self.merge_(res)

    def fine_grained_tokenize(self, tks):
        tks = tks.split(" ")
        zh_num = len([1 for c in tks if c and is_chinese(c[0])])
        if zh_num < len(tks) * 0.2:
            res = []
            for tk in tks:
                res.extend(tk.split("/"))
            return " ".join(res)

        res = []
        for tk in tks:
            if len(tk) < 3 or re.match(r"[0-9,\.-]+$", tk):
                res.append(tk)
                continue
            tkslist = []
            if len(tk) > 10:
                tkslist.append(tk)
            else:
                self.dfs_(tk, 0, [], tkslist)
            if len(tkslist) < 2:
                res.append(tk)
                continue
            stk = self.sortTks_(tkslist)[1][0]
            if len(stk) == len(tk):
                stk = tk
            else:
                if re.match(r"[a-z\.-]+$", tk):
                    for t in stk:
                        if len(t) < 3:
                            stk = tk
                            break
                    else:
                        stk = " ".join(stk)
                else:
                    stk = " ".join(stk)

            res.append(stk)

        return " ".join(self.english_normalize_(res))


def is_chinese(s):
    if s >= u'\u4e00' and s <= u'\u9fa5':
        return True
    else:
        return False


def is_number(s):
    if s >= u'\u0030' and s <= u'\u0039':
        return True
    else:
        return False


def is_alphabet(s):
    if (s >= u'\u0041' and s <= u'\u005a') or (
            s >= u'\u0061' and s <= u'\u007a'):
        return True
    else:
        return False


def naiveQie(txt):
    tks = []
    for t in txt.split(" "):
        if tks and re.match(r".*[a-zA-Z]$", tks[-1]
                            ) and re.match(r".*[a-zA-Z]$", t):
            tks.append(" ")
        tks.append(t)
    return tks


tokenizer = RagTokenizer()
tokenize = tokenizer.tokenize
fine_grained_tokenize = tokenizer.fine_grained_tokenize
tag = tokenizer.tag
freq = tokenizer.freq
loadUserDict = tokenizer.loadUserDict
addUserDict = tokenizer.addUserDict
tradi2simp = tokenizer._tradi2simp
strQ2B = tokenizer._strQ2B

if __name__ == '__main__':
    tknzr = RagTokenizer(debug=False)
    text = """
    2020年9期总第918期商业研究一、研究背景和意义2020年新春伊始 ，一场突如其来的疫情打乱了人们欢天喜地回家过年的计划 ，节日的气氛被病毒侵肆 ，1月23日，武汉打响“封城”第一枪，接着从湖北各市区乃至全国各地纷纷开 启一级应急响应 ，全国人民停工 、停产、停学,开始居家隔离 ，春节档电影除夕当天宣布延期上映 ，电影院、商场、中小商铺 、外卖、快递几乎全部停工 ，居民消费行为呈现出 “宅经济”的特点，经济社会运行受到严重冲击 。国家统计局数据显示 ，2020年1月-2月，我国社会消费品零售总额同比下降 20.5%。消费不仅是衡量 居民生活水平的重要指标 ，同时亦是我国经济增长的主要驱动 力。因此，新冠疫情在国内得到基本控制的情况下 ，国家一手紧 抓复工复产，一手紧抓内外防控 。有学者认为疫情期间居家隔离举措会影响消费者消费理念，降低整体消费水平 ，也有学者认为 ，疫情结束以后 ，居民会产生“报复性消费 ”，显著提高消费水平 ，更有认为疫情只是暂时 延缓了经济社会发展进程 ，长久看来 ，并不会对消费行为造成较 大影响。研究新冠肺炎疫情下消费者心理 、消费能力 、消费方式 、消费对象等方面的变化有助于判断未来消费行为变动趋势以及 企业应对措施 ，有助于充分挖掘国内市场 、重振居民消费 、全面建成小康社会以及实现 2020年经济社会发展目标 。二、新冠肺炎疫情特点及现状新型冠状病毒肺炎 ，简称“新冠肺炎 ”，2020年2月11日，世界卫生组织其命名为 “COVID-19 ”。
    """
    text = """
    联通云在中国品牌日绽放光芒，引领数字中国新篇章发布时间：2024年5月11日在2024年的璀璨春光中，中国迎来了又一届盛大的中国品牌日活动。这次活动以“中国品牌，世界共享；国货潮牌，共筑未来”为主题，于5月10日至14日在繁华的上海盛大举行。这场由国家发展改革委联合国务院国资委、市场监管总局、国家知识产权局共同主办的盛会，不仅是一次品牌展示的盛宴，更是中国品牌力量向世界发出的强音。中國聯通，作為数字化转型的领军者，携其旗下云计算品牌——联通云，在这场盛宴中精彩亮相，向世界展示了中国联通在智算领域的卓越成就与深远布局。一、中国品牌日：数字经济的时代号角中国品牌日自设立以来，已成为推动我国品牌建设、提升国家品牌竞争力的重要平台。今年，随着数字经济在全球范围内的蓬勃兴起，中国品牌日更是赋予了新的时代内涵。作为数字经济发展的核心驱动力，算力资源的有效配置和高效利用成为关注的焦点。中国联通以此为契机，通过联通云的亮相，向世界展示了其在算力服务领域的深厚积累与创新实践，为数字经济的高质量发展注入了强劲动力。在为期五天的展览中，中国联通展馆成为了吸引无数目光的焦点。这里不仅汇聚了联通云最新的技术成果，更通过一系列生动的场景演示，让观众亲身体验到算力技术如何深刻改变着我们的生活和工作方式。从智慧城市到智能制造，从金融科技到教育医疗，联通云以其卓越的性能和广泛的应用场景，赢得了众多参观者的高度赞誉。二、联通云：数字中国的算力引擎作为服务数字中国云计算的国家队，联通云在中国品牌日的舞台上展现了其强大的技术实力与战略视野。面对数字经济的蓬勃发展，中国联通聚焦网络强国、数字中国两大主责，明确联网通信、算网数智两大主业，以联通云为载体，全面服务数字中国“五位一体”总体布局。
    在这一过程中，联通云充分发挥算网一体的优势，成为构筑多样化算力服务的先行者。
    """
    tks = tknzr.tokenize(text)
    print(tks)
    # print(tknzr.fine_grained_tokenize(tks))
    # # huqie.addUserDict("/tmp/tmp.new.tks.dict")
    # tks = tknzr.tokenize(
    #     "哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈哈")
    # print(tknzr.fine_grained_tokenize(tks))
    # tks = tknzr.tokenize(
    #     "公开征求意见稿提出，境外投资者可使用自有人民币或外汇投资。使用外汇投资的，可通过债券持有人在香港人民币业务清算行及香港地区经批准可进入境内银行间外汇市场进行交易的境外人民币业务参加行（以下统称香港结算行）办理外汇资金兑换。香港结算行由此所产生的头寸可到境内银行间外汇市场平盘。使用外汇投资的，在其投资的债券到期或卖出后，原则上应兑换回外汇。")
    # print(tknzr.fine_grained_tokenize(tks))
    # tks = tknzr.tokenize(
    #     "多校划片就是一个小区对应多个小学初中，让买了学区房的家庭也不确定到底能上哪个学校。目的是通过这种方式为学区房降温，把就近入学落到实处。南京市长江大桥")
    # print(tknzr.fine_grained_tokenize(tks))
    # tks = tknzr.tokenize(
    #     "实际上当时他们已经将业务中心偏移到安全部门和针对政府企业的部门 Scripts are compiled and cached aaaaaaaaa")
    # print(tknzr.fine_grained_tokenize(tks))
    # tks = tknzr.tokenize("虽然我不怎么玩")
    # print(tknzr.fine_grained_tokenize(tks))
    # tks = tknzr.tokenize("蓝月亮如何在外资夹击中生存,那是全宇宙最有意思的")
    # print(tknzr.fine_grained_tokenize(tks))
    # tks = tknzr.tokenize(
    #     "涡轮增压发动机num最大功率,不像别的共享买车锁电子化的手段,我们接过来是否有意义,黄黄爱美食,不过，今天阿奇要讲到的这家农贸市场，说实话，还真蛮有特色的！不仅环境好，还打出了")
    # print(tknzr.fine_grained_tokenize(tks))
    # tks = tknzr.tokenize("这周日你去吗？这周日你有空吗？")
    # print(tknzr.fine_grained_tokenize(tks))
    # tks = tknzr.tokenize("Unity3D开发经验 测试开发工程师 c++双11双11 985 211 ")
    # print(tknzr.fine_grained_tokenize(tks))
    # tks = tknzr.tokenize(
    #     "数据分析项目经理|数据分析挖掘|数据分析方向|商品数据分析|搜索数据分析 sql python hive tableau Cocos2d-")
    # print(tknzr.fine_grained_tokenize(tks))
    # if len(sys.argv) < 2:
    #     sys.exit()
    # tknzr.DEBUG = False
    # tknzr.loadUserDict(sys.argv[1])
    # of = open(sys.argv[2], "r")
    # while True:
    #     line = of.readline()
    #     if not line:
    #         break
    #     print(tknzr.tokenize(line))
    # of.close()



#!/usr/bin/env python
# -*- coding:utf-8 _*-
"""
@author:quincy qiang
@license: Apache Licence
@file: RagApplication.py
@time: 2024/05/20
@contact: yanqiangmiffy@gamil.com
"""
import os

import pandas as pd

from trustrag.modules.citation.match_citation import MatchCitation
from trustrag.modules.document.utils import PROJECT_BASE
from trustrag.modules.generator.llm import GLM4Chat
from trustrag.modules.reranker.bge_reranker import BgeReranker, BgeRerankerConfig
from trustrag.modules.retrieval.dense_retriever import DenseRetriever, DenseRetrieverConfig


class ApplicationConfig():
    def __init__(self):
        self.retriever_config = None
        self.rerank_config = None


class WeiboRagApplication():
    def __init__(self, config):
        self.config = config
        self.retriever = DenseRetriever(self.config.retriever_config)
        self.reranker = BgeReranker(self.config.rerank_config)
        self.llm = GLM4Chat(self.config.llm_model_path)
        self.mc = MatchCitation()

    def init_vector_store(self):
        """

        """
        print("init_vector_store ... ")
        data = pd.read_json(
            os.path.join(PROJECT_BASE, 'data/docs/weibo/2024031123-2feb535f-a4b4-432d-98c6-eb1c60ddcd1f.data'),
            lines=True)
        data['text'] = data['title'] + ' ' + data['content']

        self.retriever.build_from_texts(data['text'].tolist())
        print("init_vector_store done! ")
        self.retriever.save_index(self.config.retriever_config.index_dir)

    def load_vector_store(self):
        self.retriever.load_index(self.config.retriever_config.index_path)

    def add_document(self, file_path):
        data = pd.read_json(file_path, lines=True)
        data['text'] = data['title'] + ' ' + data['content']
        for chunk in data['text'].tolist():
            self.retriever.add_text(chunk)
        print("add_document done!")

    def chat(self, question: str = '', top_k: int = 20):
        contents = self.retriever.retrieve(query=question, top_k=top_k)
        contents = self.reranker.rerank(query=question, documents=[content['text'] for content in contents])
        content = '\n'.join([content['text'] for content in contents])
        print(contents)
        response, history = self.llm.chat(question, [], content)
        result = self.mc.ground_response(
            response=response,
            evidences=[content['text'] for content in contents],
            selected_idx=[idx for idx in range(len(contents))],
            markdown=True
        )
        return result, history, contents


# if __name__ == '__main__':
#     # 修改成自己的配置！！！
#     app_config = ApplicationConfig()
#     app_config.llm_model_path = "H:/pretrained_models/llm/glm-4-9b-chat"
#
#     retriever_config = DenseRetrieverConfig(
#         model_name_or_path="H:/pretrained_models/mteb/bge-large-zh-v1.5",
#         dim=1024,
#         index_path=os.path.join(PROJECT_BASE, 'output/weibo_dense')
#     )
#     rerank_config = BgeRerankerConfig(
#         model_name_or_path="H:/pretrained_models/mteb/bge-reranker-large"
#     )
#
#     app_config.retriever_config = retriever_config
#     app_config.rerank_config = rerank_config
#     application = WeiboRagApplication(app_config)
#     # application.init_vector_store()
#     application.load_vector_store()
#     result, history, contents = application.chat("刘畅演员最近有什么活动")
#     print(result, history, contents)


import requests

class PdfParserWithMinerU:
    def __init__(self,url='http://localhost:8888/pdf_parse'):

        # 服务器URL
        self.url = url
    def parse(self,pdf_file_path,output_dir:str="output"):

        # PDF文件路径
        # pdf_file_path = 'path/to/your/file.pdf'
        print("正在基于MinerU解析pdf文件，请耐心等待，耗时时间较长。")
        # 请求参数
        params = {
            'parse_method': 'auto',
            'is_json_md_dump': 'true',
            'output_dir': output_dir
        }

        # 准备文件
        files = {
            'pdf_file': (pdf_file_path.split('/')[-1], open(pdf_file_path, 'rb'), 'application/pdf')
        }

        # 发送POST请求
        response = requests.post(self.url, params=params, files=files,timeout=2000)

        # 检查响应
        if response.status_code == 200:
            print("PDF解析成功")
            print(response.json())
            print(response.json().keys())
        else:
            print(f"错误: {response.status_code}")
            print(response.text)


if __name__ == '__main__':
    pdf_parser=PdfParserWithMinerU(url='https://aicloud.oneainexus.cn:30013/inference/aicloud-yanqiang/mineru/pdf_parse')
    pdf_file_path= '../../../data/docs/汽车操作手册.pdf'
    pdf_parser.parse(pdf_file_path)


#!/usr/bin/env python
# -*- coding:utf-8 _*-
"""
@author:quincy qiang
@license: Apache Licence
@file: utils.py
@time: 2024/06/06
@contact: yanqiangmiffy@gamil.com
@software: PyCharm
@description: coding..
"""
import os
import pathlib
import re

import tiktoken

import pathlib

# 获取当前文件所在的路径
current_path = pathlib.Path(__file__).resolve()
print(current_path)

# 找到根目录，这里假设项目的根目录为 'GoMate' 或 'GoMate_dev'
project_root = current_path
while project_root.name.lower() != 'trustrag' and project_root.name.lower() != 'trustrag_dev':
    project_root = project_root.parent
    # 如果到达根目录还没找到项目根目录，则可能路径有问题，防止死循环
    if project_root == project_root.parent:
        raise Exception("项目根目录未找到")

# 在 Windows 中输出带反斜杠的路径
project_root_str = str(project_root)
print(f"项目根目录为: {project_root_str}")


PROJECT_BASE = project_root_str
all_codecs = [
    'utf-8', 'gb2312', 'gbk', 'utf_16', 'ascii', 'big5', 'big5hkscs',
    'cp037', 'cp273', 'cp424', 'cp437',
    'cp500', 'cp720', 'cp737', 'cp775', 'cp850', 'cp852', 'cp855', 'cp856', 'cp857',
    'cp858', 'cp860', 'cp861', 'cp862', 'cp863', 'cp864', 'cp865', 'cp866', 'cp869',
    'cp874', 'cp875', 'cp932', 'cp949', 'cp950', 'cp1006', 'cp1026', 'cp1125',
    'cp1140', 'cp1250', 'cp1251', 'cp1252', 'cp1253', 'cp1254', 'cp1255', 'cp1256',
    'cp1257', 'cp1258', 'euc_jp', 'euc_jis_2004', 'euc_jisx0213', 'euc_kr',
    'gb2312', 'gb18030', 'hz', 'iso2022_jp', 'iso2022_jp_1', 'iso2022_jp_2',
    'iso2022_jp_2004', 'iso2022_jp_3', 'iso2022_jp_ext', 'iso2022_kr', 'latin_1',
    'iso8859_2', 'iso8859_3', 'iso8859_4', 'iso8859_5', 'iso8859_6', 'iso8859_7',
    'iso8859_8', 'iso8859_9', 'iso8859_10', 'iso8859_11', 'iso8859_13',
    'iso8859_14', 'iso8859_15', 'iso8859_16', 'johab', 'koi8_r', 'koi8_t', 'koi8_u',
    'kz1048', 'mac_cyrillic', 'mac_greek', 'mac_iceland', 'mac_latin2', 'mac_roman',
    'mac_turkish', 'ptcp154', 'shift_jis', 'shift_jis_2004', 'shift_jisx0213',
    'utf_32', 'utf_32_be', 'utf_32_le''utf_16_be', 'utf_16_le', 'utf_7'
]


def contains_text(text):
    # Check if the token contains at least one alphanumeric character
    return any(char.isalnum() for char in text)


def get_project_base_directory(*args):
    global PROJECT_BASE
    if PROJECT_BASE is None:
        PROJECT_BASE = os.path.abspath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                os.pardir,
                os.pardir,
            )
        )

    if args:
        return os.path.join(PROJECT_BASE, *args)
    return PROJECT_BASE


def find_codec(blob):
    global all_codecs
    for c in all_codecs:
        try:
            blob[:1024].decode(c)
            return c
        except Exception as e:
            pass
        try:
            blob.decode(c)
            return c
        except Exception as e:
            pass

    return "utf-8"


def singleton(cls, *args, **kw):
    instances = {}

    def _singleton():
        key = str(cls) + str(os.getpid())
        if key not in instances:
            instances[key] = cls(*args, **kw)
        return instances[key]

    return _singleton


def rmSpace(txt):
    txt = re.sub(r"([^a-z0-9.,]) +([^ ])", r"\1\2", txt, flags=re.IGNORECASE)
    return re.sub(r"([^ ]) +([^a-z0-9.,])", r"\1\2", txt, flags=re.IGNORECASE)


def findMaxDt(fnm):
    m = "1970-01-01 00:00:00"
    try:
        with open(fnm, "r") as f:
            while True:
                l = f.readline()
                if not l:
                    break
                l = l.strip("\n")
                if l == 'nan':
                    continue
                if l > m:
                    m = l
    except Exception as e:
        pass
    return m


def findMaxTm(fnm):
    m = 0
    try:
        with open(fnm, "r") as f:
            while True:
                l = f.readline()
                if not l:
                    break
                l = l.strip("\n")
                if l == 'nan':
                    continue
                if int(l) > m:
                    m = int(l)
    except Exception as e:
        pass
    return m


# # https://stackoverflow.com/questions/76106366/how-to-use-tiktoken-in-offline-mode-computer
# tiktoken_cache_dir = "/data/users/searchgpt/yq/GoMate/data/docs"
# os.environ["TIKTOKEN_CACHE_DIR"] = tiktoken_cache_dir
# encoder = tiktoken.get_encoding("cl100k_base")
#
#
# def num_tokens_from_string(string: str) -> int:
#     """Returns the number of tokens in a text string."""
#     num_tokens = len(encoder.encode(string))
#     return num_tokens
#
#
# def truncate(string: str, max_len: int) -> int:
#     """Returns truncated text if the length of text exceed max_len."""
#     return encoder.decode(encoder.encode(string)[:max_len])
